"""
  Simple Spider Simulator
"""
import math
from threading import Thread
from time import sleep
from datetime import timedelta

from osgar.lib.mathex import normalizeAnglePIPI
from osgar.bus import BusShutdownException


class SpiderSimulator(Thread):
    def __init__(self, config, bus):
        Thread.__init__(self)
        self.setDaemon(True)
        self.bus = bus
        self.ref_position = config['position']  # WGS84, milliseconds, [lon, lat]
        self.duration = timedelta(seconds=config.get('duration', 300))  # default 5min
        self.rel_pose = (0, 0, 0)  # rel pose to GPS position
        self.wheel_angle = 0.0  # math orientation, radians
        self.status = 0x3  # running
        self.desired_speed, self.desired_wheel_angle = None, None

    def run(self):
        try:
            self.bus.publish('status', [0x0, None])  # trigger "pump"
            self.bus.publish('position', self.get_position())
            self.bus.publish('orientation', [[0, 0, 0]]*4)  # "valid" IMU
            while True:
                timestamp, channel, data = self.bus.listen()
#                print(timestamp, channel, data)
                if timestamp > self.duration:
                    self.status = 0x0  # STOP
                if channel == 'move':
                    self.desired_speed, raw_angle = data
                    self.desired_wheel_angle = -math.pi * raw_angle / 256.0
                    self.bus.publish('status', [self.status, self.get_wheels_angles_raw()])

                    # update position at 10Hz
                    if self.status & 0x8000 == 0x0:
                        self.bus.publish('position', self.get_position())
                else:
                    assert False, channel  # unsupported channel

                dt = 0.05
                sleep(dt)  # slow down basic cycle (for now)
                self.step(dt)
        except BusShutdownException:
            pass

    def request_stop(self):
        self.bus.shutdown()

    def step(self, dt):
        if abs(self.desired_speed) > 0:
            dist = dt  # i.e. 1m/s
            if abs(self.desired_speed) <= 1.0:
                dist = 0.0  # turn in place
            x, y, heading = self.rel_pose
            a = heading + self.wheel_angle
            self.rel_pose = x + dist*math.cos(a), y + dist*math.sin(a), heading

            diff = normalizeAnglePIPI(self.desired_wheel_angle - self.wheel_angle)
            if abs(diff) > math.radians(1):
                if diff > 0:
                    corr = math.radians(10)  # 10deg/s
                else:
                    corr = math.radians(-10)
                self.wheel_angle += dt*corr
        self.status ^= 0x8000

    def get_position(self):
        "combine pose with ref GPS position"
        lon, lat = self.ref_position
        x_scale = math.cos(math.radians(lon/3600000))
        scale = 40000000/(360*3600000)
        x, y, __ = self.rel_pose
        return [int(lon + x/(scale*x_scale)), int(lat + y/scale)]

    def get_wheels_angles_raw(self):
        # Spider has 512 ticks per 360 degrees, clockwise
        raw = (-int(256 * self.wheel_angle / math.pi)) & 0x1FF
        return [raw] * 4

# vim: expandtab sw=4 ts=4




"""
  Parse Velodyne VLP-16 data
"""
import struct
import math

from osgar.node import Node

LASER_ANGLES = [-15, 1, -13, 3, -11, 5, -9, 7, -7, 9, -5, 11, -3, 13, -1, 15]
NUM_LASERS = 16


def parse_packet(data, offset_step=100):
    assert len(data) == 1206, len(data)
    assert offset_step % 100 == 0, offset_step  # must be divisible by 100
    timestamp, factory = struct.unpack_from("<IH", data, offset=1200)
    assert factory == 0x2237, hex(factory)  # 0x22=VLP-16, 0x37=Strongest Return
#    time = timestamp/1000000.0
#    if self.time is not None:
#        lost_packets = int(round((time - self.time)/EXPECTED_PACKET_TIME)) - 1
#    else:
#        lost_packets = 0
#    self.time = time
#    if lost_packets > 0 and (self.last_blocked is None or self.time > self.last_blocked + EXPECTED_SCAN_DURATION):
#        self.last_blocked = self.time + EXPECTED_SCAN_DURATION
#        self.scan_index += 1
#        print("DROPPED index", self.scan_index)
#    if self.last_blocked is not None and self.time < self.last_blocked:
#        return  # to catch up-to-date packets again ...

    ret = []
    for offset in range(0, 1200, offset_step):  # 100 bytes per one reading
        flag, azi = struct.unpack_from("<HH", data, offset)
        assert flag == 0xEEFF, hex(flag)
        azimuth = azi/100.0
        # H-distance (2mm step), B-reflectivity (0
        arr = struct.unpack_from('<' + "HB"*32, data, offset + 4)
        dist = []
        for i in range(NUM_LASERS):
            dist.append(arr[i*2] * 0.002)

        # so now we have azimuth and NUM_LASERS distance readings
        for d, beta_deg in zip(dist, LASER_ANGLES):
            beta = math.radians(beta_deg)
            x = d * math.cos(azimuth) * math.cos(beta)
            y = d * math.sin(azimuth) * math.cos(beta)
            z = d * math.sin(beta)
            ret.append([x, y, z])
    return ret


class Velodyne(Node):
    def __init__(self, config, bus):
        bus.register('raw', 'xyz')        
        super().__init__(config, bus)
        self.offset_step = config.get('offset_step', 200)  # skip every second packet (we need 1deg resolution input 0.4)
        assert self.offset_step % 100 == 0, self.offset_step  # must be divisible by 100

    def update(self):
        channel = super().update()
        assert channel == 'raw', channel
        self.publish('xyz', parse_packet(self.raw, offset_step=self.offset_step))

# vim: expandtab sw=4 ts=4



#!/usr/bin/python
"""
  Conversion utility from log files to pcap format
  usage:
       ./log2pcap.py <input file(s)> <output directory>
"""
import sys
import os
from binascii import unhexlify

from osgar.logger import LogReader, lookup_stream_id
from osgar.lib.serialize import deserialize

# Notes:
#   Source https://github.com/kisom/pypcapfile was used for parsing original
# pcap files. There is 24byte header and each packet has 16byte header with
# timestamp and packet length followed by IP/UDP header and data.
#
# The data can be then viewer via VeloView:
#    http://www.paraview.org/Wiki/VeloView


FILE_HEADER = 'd4c3b2a1020004000000000000000000ffff000001000000'
PACKET_HEADER = '715fb15517bb0a00e0040000e0040000'
IP_HEADER = 'ffffffffffff6076880000000800450004d200004000ff11b4aac0a801c8ffffffff0940094004be0000'


def log2pcap(input_filepath, output_dir):
    """
       Extract from logfile Velodyne UPD packets and save them in 'pcap' format
       undestandable by Wireshark and VeloView.
    """
    output_filepath = os.path.join(output_dir, os.path.basename(input_filepath))
    assert output_filepath.endswith('.log'), output_filepath
    output_filepath = output_filepath[:-4] + '.pcap'
    print(output_filepath)

    only_stream = lookup_stream_id(input_filepath, 'velodyne.raw')
    with LogReader(input_filepath, only_stream_id=only_stream) as log, open(output_filepath, 'wb') as out:
        out.write(unhexlify(FILE_HEADER))
        for timestamp, stream_id, data in log:
            packet = deserialize(data)
            assert len(packet) == 1206, len(packet)

            out.write(unhexlify(PACKET_HEADER))  # TODO revise timestamps
            out.write(unhexlify(IP_HEADER))
            out.write(packet)


def main():
    if len(sys.argv) < 3:
        print(__doc__)
        sys.exit(2)

    output_dir = sys.argv[-1]
    for filename in sys.argv[1:-1]:
        log2pcap(filename, output_dir)


if __name__ == "__main__":
    main()

# vim: expandtab sw=4 ts=4 




#!/usr/bin/env python2

from __future__ import print_function

import math
import socket
import sys
import threading
import time
import operator
import functools

import rospy
import rostopic
import zmq
import numpy as np

from sensor_msgs.msg import Imu, LaserScan, CompressedImage, Image, PointCloud2
from nav_msgs.msg import Odometry
from std_msgs.msg import Empty, Int32
from sensor_msgs.msg import BatteryState, FluidPressure

sys.path.append("/osgar-ws/src/osgar/osgar/lib")
import serialize as osgar_serialize

def py3round(f):
    if abs(round(f) - f) == 0.5:
        return int(2.0 * round(f / 2.0))
    return int(round(f))


class Bus:
    def __init__(self):
        self.lock = threading.Lock()
        context = zmq.Context.instance()
        self.push = context.socket(zmq.PUSH)
        self.push.setsockopt(zmq.LINGER, 100)  # milliseconds
        self.push.bind('tcp://*:5565')
        self.pull = context.socket(zmq.PULL)
        self.pull.LINGER = 100
        self.pull.RCVTIMEO = 100
        self.pull.bind('tcp://*:5566')

    def register(self, *outputs):
        pass

    def publish(self, channel, data):
        raw = osgar_serialize.serialize(data)
        with self.lock:
            self.push.send_multipart([channel, raw])

    def listen(self):
        while not rospy.is_shutdown():
            try:
                channel, bytes_data = self.pull.recv_multipart()
                data = osgar_serialize.deserialize(bytes_data)
                return channel.decode('ascii'), data
            except zmq.ZMQError as e:
                if e.errno != zmq.EAGAIN:
                    rospy.logerr("zmq error")
                    sys.exit("zmq error")
        rospy.loginfo("done")
        sys.exit()


class main:
    def __init__(self, robot_name, robot_config, robot_is_marsupial):
        rospy.init_node('cloudsim2osgar', log_level=rospy.DEBUG)
        self.bus = Bus()

        # common topics
        topics = [
            ('/' + robot_name + '/battery_state', BatteryState, self.battery_state, ('battery_state',)),
            ('/subt/score', Int32, self.score, ('score',)),
        ]

        publishers = {}

        # configuration specific topics
        robot_description = rospy.get_param("/{}/robot_description".format(robot_name))
        if robot_config == "ROBOTIKA_X2_SENSOR_CONFIG_1":
            rospy.loginfo("robotika x2")
        elif robot_config == "SSCI_X4_SENSOR_CONFIG_2":
            rospy.loginfo("ssci drone")
            topics.append(('/' + robot_name + '/top_scan', LaserScan, self.top_scan, ('top_scan',)))
            topics.append(('/' + robot_name + '/bottom_scan', LaserScan, self.bottom_scan, ('bottom_scan',)))
            topics.append(('/' + robot_name + '/odom_fused', Odometry, self.odom_fused, ('pose3d',)))
            topics.append(('/' + robot_name + '/air_pressure', FluidPressure, self.air_pressure, ('air_pressure',)))
            if robot_is_marsupial == 'true':
                rospy.loginfo("X4 is marsupial")
                publishers['detach'] = rospy.Publisher('/' + robot_name + '/detach', Empty, queue_size=1)
        elif robot_config == "TEAMBASE":
            rospy.loginfo("teambase")
        elif robot_config.startswith("ROBOTIKA_FREYJA_SENSOR_CONFIG"):
            if robot_config.endswith("_2"):
                rospy.loginfo("freya 2 (with comms beacons)")
                publishers['deploy'] = rospy.Publisher('/' + robot_name + '/breadcrumb/deploy', Empty, queue_size=1)
            else:
                rospy.loginfo("freya 1 (basic)")
            topics.append(('/' + robot_name + '/odom_fused', Odometry, self.odom_fused, ('pose3d',)))
            topics.append(('/' + robot_name + '/rgbd_front/image_raw/compressed', CompressedImage, self.image_front, ('image_front',)))
            topics.append(('/' + robot_name + '/rgbd_rear/image_raw/compressed', CompressedImage, self.image_rear, ('image_rear',)))
            topics.append(('/' + robot_name + '/scan_front', LaserScan, self.scan_front, ('scan_front',)))
            topics.append(('/' + robot_name + '/scan_rear', LaserScan, self.scan_rear, ('scan_rear',)))
            topics.append(('/' + robot_name + '/rgbd_front/depth', Image, self.depth_front, ('depth_front',)))
            topics.append(('/' + robot_name + '/rgbd_rear/depth', Image, self.depth_rear, ('depth_rear',)))
        elif robot_config.startswith("ROBOTIKA_KLOUBAK_SENSOR_CONFIG"):
            if robot_config.endswith("_2"):
                rospy.loginfo("k2 2 (with comms beacons)")
                publishers['deploy'] = rospy.Publisher('/' + robot_name + '/breadcrumb/deploy', Empty, queue_size=1)
            else:
                rospy.loginfo("k2 1 (basic)")
            topics.append(('/' + robot_name + '/odom_fused', Odometry, self.odom_fused, ('pose3d',)))
        elif robot_config.startswith("EXPLORER_R2_SENSOR_CONFIG"):
            if robot_config.endswith("_2"):
                rospy.loginfo("explorer R2 #2 (with comms beacons)")
                publishers['deploy'] = rospy.Publisher('/' + robot_name + '/breadcrumb/deploy', Empty, queue_size=1)
            else:
                rospy.loginfo("explorer R2 #1 (basic)")
            topics.append(('/' + robot_name + '/rs_front/color/image/compressed', CompressedImage, self.image_front, ('image_front',)))
            topics.append(('/' + robot_name + '/rs_front/depth/image', Image, self.depth_front, ('depth_front',)))
            topics.append(('/' + robot_name + '/points', PointCloud2, self.points, ('points',)))
        else:
            rospy.logerr("unknown configuration")
            return

        if robot_config.startswith("ROBOTIKA_KLOUBAK_SENSOR_CONFIG"):
            topics.append(('/' + robot_name + '/imu/front/data', Imu, self.imu, ('rot', 'acc', 'orientation')))
        else:
            topics.append(('/' + robot_name + '/imu/data', Imu, self.imu, ('rot', 'acc', 'orientation')))

        outputs = functools.reduce(operator.add, (t[-1] for t in topics))
        self.bus.register(outputs)

        for name, type, handler, _ in topics:
            rospy.loginfo("waiting for {}".format(name))
            rospy.wait_for_message(name, type)
            setattr(self, handler.__name__+"_count", 0)
            rospy.Subscriber(name, type, handler)

        # main thread receives data from osgar and sends it to ROS
        while True:
            channel, data = self.bus.listen()
            rospy.logdebug("receiving: {} {}".format(channel, data))
            # switch on channel to feed different ROS publishers
            if channel in publishers:
                publishers[channel].publish(*data)  # just guessing for Empty, where ([],) is wrong
            else:
                rospy.loginfo("ignoring: {} {}".format(channel, data))

    def imu(self, msg):
        self.imu_count += 1
        rospy.loginfo_throttle(10, "imu callback: {}".format(self.imu_count))
        acc = [msg.linear_acceleration.x, msg.linear_acceleration.y, msg.linear_acceleration.z]
        orientation = [msg.orientation.x, msg.orientation.y, msg.orientation.z, msg.orientation.w]

        # copy & paste from rosmsg
        q0, q1, q2, q3 = orientation  # quaternion
        x = math.atan2(2 * (q0 * q1 + q2 * q3), 1 - 2 * (q1 * q1 + q2 * q2))
        y = math.asin(2 * (q0 * q2 - q3 * q1))
        z = math.atan2(2 * (q0 * q3 + q1 * q2), 1 - 2 * (q2 * q2 + q3 * q3))
        rot = [x, y, z]
        # end of copy & paste from rosmsg

        self.bus.publish('rot', [py3round(math.degrees(angle) * 100) for angle in rot])
        self.bus.publish('acc', [py3round(x * 1000) for x in acc])
        self.bus.publish('orientation', orientation)
        # preliminary suggestion for combined message
        #angular_velocity = [msg.angular_velocity.x, msg.angular_velocity.y, msg.angular_velocity.z]
        #data = [
        #    msg.header.stamp.to_nsec()/1000000, # time in milliseconds
        #    orientation,
        #    angular_velocity,
        #    acc,
        #]
        #self.bus.publish('imu', data)

    def top_scan(self, msg):
        self.top_scan_count += 1
        rospy.loginfo_throttle(10, "top_scan callback: {}".format(self.top_scan_count))
        self.bus.publish('top_scan', msg.ranges)

    def bottom_scan(self, msg):
        self.bottom_scan_count += 1
        rospy.loginfo_throttle(10, "bottom_scan callback: {}".format(self.bottom_scan_count))
        self.bus.publish('bottom_scan', msg.ranges)

    def odom_fused(self, msg):
        self.odom_fused_count += 1
        rospy.loginfo_throttle(10, "odom_fused callback: {}".format(self.odom_fused_count))

        position = msg.pose.pose.position
        orientation = msg.pose.pose.orientation

        xyz = [position.x, position.y, position.z]
        orientation = [orientation.x, orientation.y, orientation.z, orientation.w]

        self.bus.publish('pose3d', [xyz, orientation])

    def battery_state(self, msg):
        self.battery_state_count += 1
        rospy.loginfo_throttle(10, "battery_state callback: {}".format(self.battery_state_count))
        if getattr(self, 'last_battery_secs', None) != msg.header.stamp.secs:
            self.bus.publish("battery_state", msg.percentage)
            self.last_battery_secs = msg.header.stamp.secs

    def score(self, msg):
        self.score_count += 1
        rospy.loginfo_throttle(10, "score callback: {}".format(self.score_count))
        self.bus.publish("score", msg.data)

    def air_pressure(self, msg):
        self.air_pressure_count += 1
        rospy.loginfo_throttle(10, "air_pressure callback: {}".format(self.air_pressure_count))
        self.bus.publish('air_pressure', msg.fluid_pressure)

    def image_front(self, msg):
        self.image_front_count += 1
        rospy.loginfo_throttle(10, "image_front callback: {}".format(self.image_front_count))
        self.bus.publish('image_front', msg.data)

    def image_rear(self, msg):
        self.image_rear_count += 1
        rospy.loginfo_throttle(10, "image_rear callback: {}".format(self.image_rear_count))
        self.bus.publish('image_rear', msg.data)

    def scan_front(self, msg):
        self.scan_front_count += 1
        rospy.loginfo_throttle(10, "scan_front callback: {}".format(self.scan_front_count))
        scan = [int(x * 1000) if msg.range_min < x < msg.range_max else 0 for x in msg.ranges]
        self.bus.publish('scan_front', scan)

    def scan_rear(self, msg):
        self.scan_rear_count += 1
        rospy.loginfo_throttle(10, "scan_rear callback: {}".format(self.scan_rear_count))
        scan = [int(x * 1000) if msg.range_min < x < msg.range_max else 0 for x in msg.ranges]
        self.bus.publish('scan_rear', scan)

    def convert_depth(self, msg):
        assert msg.encoding == '32FC1', msg.encoding  # unsupported encoding
        # depth is array of floats, OSGAR uses uint16 in millimeters
        # cut min & max (-inf and inf are used for clipping)
        arr = np.frombuffer(msg.data, dtype=np.dtype('f')) * 1000
        arr = np.clip(arr, 1, 0xFFFF)
        arr = np.ndarray.astype(arr, dtype=np.dtype('H'))
        return np.array(arr).reshape((msg.height, msg.width))

    def depth_front(self, msg):
        self.depth_front_count += 1
        rospy.loginfo_throttle(10, "depth_front callback: {}".format(self.depth_front_count))
        self.bus.publish('depth_front', self.convert_depth(msg))

    def depth_rear(self, msg):
        self.depth_rear_count += 1
        rospy.loginfo_throttle(10, "depth_rear callback: {}".format(self.depth_rear_count))
        self.bus.publish('depth_rear', self.convert_depth(msg))

    def convert_points(self, msg):
        # accept only Velodyne VLC-16 (for the ver0)
        assert msg.height == 16, msg.height
        assert msg.width == 10000, msg.width
        assert msg.point_step == 32, msg.point_step
        assert msg.row_step == 320000, msg.row_step
        arr = np.frombuffer(msg.data, dtype=np.float32)
        points3d = arr.reshape((msg.height, msg.width, 8))[:, :, 0:3]  # keep only (x, y, z)
        return points3d[:, ::10, :]  # downsample to everh 10th

    def points(self, msg):
        self.points_count += 1
        rospy.loginfo_throttle(10, "points callback: {}".format(self.points_count))
        self.bus.publish('points', self.convert_points(msg))


if __name__ == '__main__':
    if len(sys.argv) < 4:
        print("need robot name, config and is_marsupial as arguments", file=sys.stderr)
        sys.exit(2)
    try:
        main(sys.argv[1], sys.argv[2], sys.argv[3])
    except rospy.exceptions.ROSInterruptException:
        rospy.loginfo("shutdown")




from .unittest import TestCase

import numpy as np


class Test(TestCase):

    def test_numpy(self):
        template = [1, 2, 3]
        a = np.asarray(template)
        b = np.asarray(template)
        self.assertEqual(a, b)
        self.assertNumpyEqual(a, b)

        with self.assertRaises(AssertionError):
            self.assertNumpyEqual(a, [])
        with self.assertRaises(AssertionError):
            self.assertNumpyEqual(a, template)

    def test_pose3d(self):
        a = [[0,0,0], [0,0,0,1]]
        with self.assertRaises(AssertionError):
            self.assertPose3dEqual([], [])

        self.assertPose3dEqual(a, a)

    def test_list(self):
        self.assertListAlmostEqual([1], [1])
        self.assertListAlmostEqual([1, 2], [1, 2])
        self.assertListAlmostEqual([1.1, 2], [1.1, 2])
        self.assertListAlmostEqual([1.10000001, 2], [1.1, 2])

        with self.assertRaises(AssertionError):
            self.assertListAlmostEqual([1.1000001, 2], [1.1, 2])
        with self.assertRaises(AssertionError):
            self.assertListAlmostEqual([1, 2], [1, 3])
        with self.assertRaises(AssertionError):
            self.assertListAlmostEqual([1, 2], [1, 2, 3])



"""
  Simple IMU data parsing (VectorNav VN-100)
"""

from threading import Thread

from osgar.bus import BusShutdownException


def parse_line(line):
    """
    Parse $VNYMR message:
        Yaw  float  deg  Calculated attitude heading angle in degrees.
        Pitch  float  deg  Calculated attitude pitch angle in degrees.
        Roll  float  deg  Calculated attitude roll angle in degrees.
        MagX  float  Gauss  Compensated magnetometer measurement in x-axis.
        MagY  float  Gauss  Compensated magnetometer measurement in y-axis.
        MagZ  float  Gauss  Compensated magnetometer measurement in z-axis.
        AccelX  float  m/s^2 Compensated accelerometer measurement in x-axis.
        AccelY  float  m/s^2 Compensated accelerometer measurement in y-axis.
        AccelZ  float  m/s^2 Compensated accelerometer measurement in z-axis.
        GyroX  float  rad/s  Compensated angular rate in x-axis.
        GyroY  float  rad/s  Compensated angular rate in y-axis.
        GyroZ  float  rad/s  Compensated angular rate in z-axis.
    """
    assert line.startswith(b'$VNYMR'), line
    assert b'*' in line, line
    s = line.split(b'*')[0].split(b',')
    assert len(s) == 13, s
    arr = [float(x) for x in s[1:]]
    return [arr[:3], arr[3:6], arr[6:9], arr[9:]]


class IMU(Thread):
    def __init__(self, config, bus):
        bus.register('orientation', 'rotation')
        Thread.__init__(self)
        self.setDaemon(True)

        self.offset = config.get("offset", [0, 0, 0])  # rotation offset in 1/100th of degree

        self.bus = bus
        self.buf = b''

    # Copy & Paste from gps.py - refactor!
    @staticmethod
    def split_buffer(data):
        start = data.find(b'$')
        if start < 0:
            return data, b''
        end = data[start:-2].find(b'*')
        if end < 0:
            return data, b''
        return data[start+end+3:], data[start:start+end+3]

    def process_packet(self, line):
        if line.startswith(b'$VNYMR'):
            result = parse_line(line)
            return result
        return None

    def process_gen(self, data):
        self.buf, packet = self.split_buffer(self.buf + data)
        while len(packet) > 0:
            ret = self.process_packet(packet)
            if ret is not None:
                yield ret
            # now process only existing (remaining) buffer
            self.buf, packet = self.split_buffer(self.buf)  

    def run(self):
        try:
            while True:
                packet = self.bus.listen()
                dt, __, data = packet
                for out in self.process_gen(data):
                    assert out is not None
                    self.bus.publish('orientation', out)
                    # publish separately yaw, pitch and roll in 1/100th deg
                    yaw, pitch, roll = out[0]
                    # The  VN-100  uses a right-handed coordinate system. A positive yaw angle
                    # is defined as a positive righthanded rotation around the Z-axis. (pointing down)
                    yaw = 90.0 - yaw
                    if yaw < -180:
                        yaw += 360
                    if yaw > 180:
                        yaw -= 360
                    angles = [int(round(x * 100)) for x in [yaw, pitch, roll]]
                    angles = [a + b for a, b in zip(angles, self.offset)]  # correct for offset
                    self.bus.publish('rotation', angles)
        except BusShutdownException:
            pass

    def request_stop(self):
        self.bus.shutdown()


if __name__ == "__main__":
    import io
    import sys
    import matplotlib.pyplot as plt
    
    arr = []
    for line in io.open(sys.argv[1]):
        angle, mag, acc, gyro = parse_line(line)
        arr.append(angle)

    plt.plot(arr, 'o-', linewidth=2)
    plt.show()

# vim: expandtab sw=4 ts=4



import unittest
import math

import numpy as np

from subt.depth2scan import vertical_step, monotize, find_step


class Depth2ScanTest(unittest.TestCase):

    def test_vertical_step(self):
        depth = np.zeros((360, 640), dtype=np.uint16)
        self.assertIsNone(vertical_step(depth))

        depth[:, 320] = np.arange(0, 4*360, 4)
        #self.assertIsNone(vertical_step(depth))

        depth = np.zeros((360, 640), dtype=np.uint16)
        depth[:,:] = 200
        for i in range(180, 250):
            depth[i, 320] = 400
        self.assertEqual(vertical_step(depth), 200)  # TODO better example

    def test_monotize(self):
        a = np.zeros(10, dtype=np.int32)
        np.testing.assert_equal(monotize(a), a)
        a[7] = 13
        self.assertEqual(monotize(a)[-1], 13)

    def test_find_step(self):
        self.assertIsNone(find_step(np.zeros(250, dtype=np.int32)))
        
        arr = np.zeros(250, dtype=np.int32)
        arr[100:] = 200  # wooden crate 20cm high
        self.assertEqual(find_step(arr), 100)

# vim: expandtab sw=4 ts=4




"""
  Recorder for given nodes configuration
"""

import argparse
import sys
import os
import signal
import threading
import logging

from osgar.logger import LogWriter
from osgar.lib.config import config_load, get_class_by_name
from osgar.bus import Bus

g_logger = logging.getLogger(__name__)

class Recorder:
    def __init__(self, config, logger):
        self.stop_requested = threading.Event()
        self.sigint_received = False
        self.modules = {}

        self.bus = Bus(logger)
        for module_name, module_config in config['modules'].items():
            klass = get_class_by_name(module_config['driver'])
            self.modules[module_name] = klass(module_config.get('init', {}), bus=self.bus.handle(module_name))

        for sender, receiver in config['links']:
            self.bus.connect(sender, receiver, self.modules)

        signal.signal(signal.SIGINT, self.request_stop)
        g_logger.info("SIGINT handler installed")

    def __enter__(self):
        for module in self.modules.values():
            module.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.request_stop()
        for module in self.modules.values():
            module.join(1)
        for name, max_delay, timestamp in self.bus.delays():
            g_logger.error(f"{name:>12}: maximum delay of {max_delay} at {timestamp}")
        for t in threading.enumerate():
            if t != threading.current_thread():
                g_logger.error(f'thread {repr(t.name)} still running!')
                g_logger.error(f'    class: {t.__class__.__module__}.{t.__class__.__name__}')
                g_logger.error(f'    target: {t._target}')
        if self.sigint_received:
            g_logger.info("committing suicide by SIGINT")
            signal.signal(signal.SIGINT, signal.SIG_DFL)
            os.kill(os.getpid(), signal.SIGINT)

    def request_stop(self, signum=None, frame=None): # pylint: disable=unused-argument
        if signum == signal.SIGINT:
            self.sigint_received = True
        if self.stop_requested.is_set():
            return
        for module in self.modules.values():
            module.request_stop()
        self.stop_requested.set()


def record(config, log_prefix, log_filename=None, duration_sec=None):
    with LogWriter(prefix=log_prefix, filename=log_filename, note=str(sys.argv)) as log:
        log.write(0, bytes(str(config), 'ascii'))
        g_logger.info(log.filename)
        with Recorder(config=config['robot'], logger=log) as recorder:
            if 'app' in recorder.modules:
                app = recorder.modules['app']
                app.join(duration_sec)
            else:
                recorder.stop_requested.wait(duration_sec)


def main(record=record):
    import logging
    logging.basicConfig(
        level=logging.DEBUG,
        format='%(asctime)s %(name)-16s %(levelname)-8s %(message)s',
    )
    parser = argparse.ArgumentParser(description='Record run on real HW with given configuration')
    parser.add_argument('config', nargs='+', help='configuration file')
    parser.add_argument('--note', help='add description')
    parser.add_argument('--duration', help='recording duration (sec), default infinite', type=float)
    parser.add_argument('--log', help='force record log filename')
    parser.add_argument('--application', help='import string to application', default=None)
    args = parser.parse_args()

    prefix = os.path.basename(args.config[0]).split('.')[0] + '-'
    cfg = config_load(*args.config, application=args.application)
    record(cfg, log_prefix=prefix, duration_sec=args.duration, log_filename=args.log)

if __name__ == "__main__":
    main()

# vim: expandtab sw=4 ts=4



import cv2
import numpy as np
import torch

ARTIFACT_CATEGORIES = dict((name, artf_id) for artf_id, name in enumerate([
    'nothing', 'backpack', 'phone', 'survivor', 'vent', 'robot', 'helmet',
    'rope', 'breadcrumb'
    #'drill', 'toolbox', 'fire_extinguisher', 'electrical_box', 'radio', 'valve'
]))


class SubtDataset(torch.utils.data.Dataset):
    def __init__(self,
                 labels_config,
                 downscale,
                 receptive_field,
                 transform=None,
                 num_transforms=4,
                 a_few_negatives=200,
                 nothing_stride=10):
        import json
        import os
        import PIL

        super(SubtDataset, self).__init__()
        if transform is None:
            transform = torchvision.transforms.CenterCrop(receptive_field)
            num_transforms = 1

        with open(labels_config, 'r') as labels_file:
            labels = json.loads(labels_file.read())

        self.patches = []
        self.artifacts = []
        self.origins = []
        img_root = os.path.dirname(labels_config)

        blacklist = [
            ('backpack', '../virtual/backpack/backpack3.jpg'),  # ?
            ('backpack',
             '../virtual/backpack/rgbd/ji-ver54-saveX-0021.jpg'),  # ?
            ('backpack',
             '../virtual/survivor/rgbd/ji-ver54-saveX-0022.jpg'),  # ?
            ('backpack',
             '../virtual/survivor/rgbd/ji-ver54-saveX-0023.jpg'),  # ?
            ('phone', '../virtual/phone/saveX-0001.jpg'),
            ('phone', '../virtual/phone/saveX-0002.jpg'),
            ('phone', '../virtual/phone/saveX-0003.jpg'),
            ('phone', '../virtual/phone/saveX-0004.jpg'),
            ('phone', '../virtual/phone/saveX-0005.jpg'),
            ('phone', '../virtual/phone/saveX-0006.jpg'),
            ('phone', '../virtual/phone/saveX-0007.jpg'),
            ('phone', '../virtual/phone/saveX-0008.jpg'),
            ('phone', '../virtual/phone/saveX-0009.jpg'),
            ('phone', '../virtual/phone/saveX-0010.jpg'),
            ('phone', '../virtual/phone/saveX-0011.jpg'),
            ('phone', '../virtual/phone/saveX-0012.jpg'),
            ('phone', '../virtual/phone/saveX-0013.jpg'),
            ('phone', '../virtual/phone/saveX-0014.jpg'),
            ('phone', '../virtual/phone/saveX-0015.jpg'),
            ('phone', '../virtual/phone/saveX-0016.jpg'),
            ('phone', '../virtual/phone/saveX-0017.jpg'),
            ('phone', '../virtual/phone/saveX-0018.jpg'),
            ('phone', '../virtual/phone/saveX-0019.jpg'),
            ('phone', '../virtual/phone/saveX-0020.jpg'),
            ('phone', '../virtual/phone/saveX-0021.jpg'),
            ('phone', '../virtual/phone/saveX-0022.jpg'),
            ('phone', '../virtual/phone/saveX-0023.jpg'),
            ('phone', '../virtual/phone/saveX-0024.jpg'),
            ('phone', '../virtual/phone/saveX-0025.jpg'),
            ('phone', '../virtual/phone/saveX-0026.jpg'),
            ('phone', '../virtual/phone/saveX-0027.jpg'),
            ('phone', '../virtual/phone/saveX-0028.jpg'),
            ('phone', '../virtual/phone/saveX-0029.jpg'),
            ('phone', '../virtual/phone/saveX-0030.jpg'),
            ('phone', '../virtual/phone/saveX-0031.jpg'),
            ('phone', '../virtual/phone/saveX-0032.jpg'),
            #('phone', '../virtual/errors/ji-ver54-saveX-0014.jpg'),  # ?
            ('phone', '../virtual/errors/ji-ver54-saveX-0018.jpg'),  # ?
            ('survivor', '../virtual/survivor/005.jpg'),
            ('survivor', '../virtual/survivor/rgbd/ji-ver54-saveX-0023.jpg'),
            ('survivor', '../virtual/survivor/survivor-far.jpg')
        ]
        for label in labels.values():
            img_path = os.path.join(img_root, label['filename'])
            img = cv2.imread(img_path, cv2.IMREAD_COLOR)
            img = img[::downscale, ::downscale, :]
            h, w, _ = img.shape

            pts = []
            for region in label['regions']:
                shape = region['shape_attributes']
                x = shape['x'] // downscale
                y = shape['y'] // downscale
                width = shape['width'] // downscale
                height = shape['height'] // downscale
                artifact = region['region_attributes']['artifact']
                if artifact not in ARTIFACT_CATEGORIES:
                    continue
                if (artifact, label['filename']) in blacklist:
                    continue

                # Focus on the center of the artifact.
                pts.append((artifact, x + width / 2, y + height / 2))

                # Everything in a 'nothing' box is nothing.
                if artifact == 'nothing':
                    for v in range(y, y + height, nothing_stride):
                        for u in range(x, x + width, nothing_stride):
                            pts.append((artifact, u, v))

            # Add a few random points outside of bounding boxes as other
            # examples of 'nothing'.
            rnd_xs = np.random.default_rng().integers(0, img.shape[1],
                                                      a_few_negatives)
            rnd_ys = np.random.default_rng().integers(0, img.shape[0],
                                                      a_few_negatives)
            for (rnd_x, rnd_y) in zip(rnd_xs, rnd_ys):
                is_outside = True
                for region in label['regions']:
                    shape = region['shape_attributes']
                    x = shape['x'] // downscale
                    y = shape['y'] // downscale
                    width = shape['width'] // downscale
                    height = shape['height'] // downscale

                    if (x <= rnd_x < x + width) and (y <= rnd_y < y + height):
                        is_outside = False
                        break

                if is_outside:
                    pts.append(('nothing', rnd_x, rnd_y))

            for artifact, cx, cy in pts:
                # Giving ourselves some slack for transforms
                ltx = int(cx - receptive_field)
                lty = int(cy - receptive_field)
                if ltx < 0 or lty < 0:
                    continue
                rbx = ltx + receptive_field * 2
                rby = lty + receptive_field * 2
                if rbx > w or rby > h:
                    continue
                patch = img[lty:rby, ltx:rbx, :]
                assert (patch.shape == (receptive_field * 2,
                                        receptive_field * 2, 3))

                patch = PIL.Image.fromarray(
                    cv2.cvtColor(patch, cv2.COLOR_BGR2RGB))
                for _ in range(num_transforms):
                    patch_bgr = cv2.cvtColor(np.array(transform(patch)),
                                             cv2.COLOR_RGB2BGR)
                    self.patches.append(patch_bgr)
                    self.artifacts.append(artifact)
                    self.origins.append(label['filename'])

    def __len__(self):
        return len(self.patches)

    def __getitem__(self, idx):
        return (self.patches[idx], ARTIFACT_CATEGORIES[self.artifacts[idx]],
                self.origins[idx])


class ChannelFirst(torch.nn.Module):
    def forward(self, x):
        # NHWC -> NCHW
        return x.permute(0, 3, 1, 2)


class ChannelLast(torch.nn.Module):
    def forward(self, x):
        # NCHW -> NHWC
        return x.permute(0, 2, 3, 1)


class Downscale(torch.nn.Module):
    def __init__(self, factor):
        super(Downscale, self).__init__()
        self.factor = factor

    def forward(self, x):
        return x[:, ::self.factor, ::self.factor, :]


class NormalizeImage(torch.nn.Module):
    def forward(self, x):
        return (x - 127.) / 128.


class Swish(torch.nn.Module):
    def forward(self, x):
        return x * torch.sigmoid(x)


class HSwish(torch.nn.Module):
    def forward(self, x):
        return x * torch.nn.functional.relu6(x + 3) / 6


class MaybeSoftMax(torch.nn.Module):
    def forward(self, x):
        if self.training:
            return x
        else:
            return torch.nn.functional.softmax(x, dim=1)


def activation_fn(fn):
    return {
        'relu': torch.nn.ReLU,
        'gelu': torch.nn.GELU,
        'swish': Swish,
        'hswish': HSwish
    }[fn]()


class MDNet0(torch.nn.Sequential):
    def __init__(self, downscale, embedding_size_0, embedding_size_1,
                 activation, dropout_probability, receptive_field,
                 num_categories):
        super(MDNet0, self).__init__(
            Downscale(downscale), NormalizeImage(), ChannelFirst(),
            torch.nn.Conv2d(3, embedding_size_0, kernel_size=1),
            torch.nn.BatchNorm2d(embedding_size_0), activation_fn(activation),
            torch.nn.Conv2d(embedding_size_0, embedding_size_0, kernel_size=1),
            torch.nn.BatchNorm2d(embedding_size_0), activation_fn(activation),
            torch.nn.Dropout2d(dropout_probability),
            torch.nn.Conv2d(embedding_size_0,
                            embedding_size_1,
                            kernel_size=receptive_field,
                            padding=receptive_field // 2),
            torch.nn.BatchNorm2d(embedding_size_1), activation_fn(activation),
            torch.nn.Conv2d(embedding_size_1, num_categories, kernel_size=1),
            MaybeSoftMax(), ChannelLast())


class MBConv(torch.nn.Module):
    ''' A block from MobileNet V2. '''
    def __init__(self,
                 input_channels,
                 expansion_factor,
                 activation,
                 kernel=3,
                 stride=1):
        super(MBConv, self).__init__()
        expanded_channels = expansion_factor * input_channels
        self.cconv0 = torch.nn.Conv2d(input_channels,
                                      expanded_channels,
                                      kernel_size=1)
        self.bn0 = torch.nn.BatchNorm2d(expanded_channels)
        self.dconv = torch.nn.Conv2d(expanded_channels,
                                     expanded_channels,
                                     kernel_size=kernel,
                                     padding=kernel // 2,
                                     groups=expanded_channels)
        self.bn1 = torch.nn.BatchNorm2d(expanded_channels)
        self.cconv1 = torch.nn.Conv2d(expanded_channels,
                                      input_channels,
                                      kernel_size=1)
        self.bn2 = torch.nn.BatchNorm2d(input_channels)
        self.activation = activation_fn(activation)

    def forward(self, x):
        x0 = x
        x = self.cconv0(x)
        x = self.bn0(x)
        x = self.activation(x)
        x - self.dconv(x)
        x = self.bn1(x)
        x = self.activation(x)
        x = self.cconv1(x)
        x = self.bn2(x)
        x += x0
        return x


class MBNet0(torch.nn.Sequential):
    def __init__(self, downscale, dropout_probability, embedding_size,
                 activation, convolution_layers, embedding_expansion_factor,
                 kernel_size, num_categories):
        conv_layers = [
            MBConv(embedding_size, embedding_expansion_factor, activation,
                   kernel_size) for _ in range(convolution_layers)
        ]
        super(MBNet0, self).__init__(
            Downscale(downscale), NormalizeImage(), ChannelFirst(),
            torch.nn.Dropout2d(dropout_probability),
            torch.nn.Conv2d(3, embedding_size, 3, 1, 3 // 2), *conv_layers,
            torch.nn.Conv2d(embedding_size, num_categories, kernel_size=1),
            MaybeSoftMax(), ChannelLast())


def create_model(model_config):
    # Other model architectures may be supported in future.
    assert (model_config['type'] in ['mdnet0', 'mbnet0'])
    if model_config['type'] == 'mdnet0':
        model = MDNet0(model_config['downscale'],
                       model_config['embedding-size-0'],
                       model_config['embedding-size-1'],
                       model_config['activation'],
                       model_config['dropout-probability'],
                       model_config['receptive-field'],
                       len(model_config['categories']))
    elif model_config['type'] == 'mbnet0':
        model = MBNet0(model_config['downscale'],
                       model_config['dropout-probability'],
                       model_config['embedding-size'],
                       model_config['activation'],
                       model_config['convolution-layers'],
                       model_config['embedding-expansion-factor'],
                       model_config['kernel-size'],
                       len(model_config['categories']))
    return model


def load_model(path, device):
    model_params = torch.load(path, map_location=torch.device(device))
    model_type = model_params.get('type', 'mdnet0')
    model = create_model(model_params)
    model = model.float().to(device)
    model.load_state_dict(model_params['model'])
    model.eval()

    return model, model_params['categories']


def save_model(model_config, model, path):
    # Other model architectures may be supported in future.
    assert (model_config['type'] in ['mdnet0', 'mbnet0'])
    if model_config['type'] == 'mdnet0':
        torch.save(
            {
                'type': 'mdnet0',
                'downscale': model_config['downscale'],
                'dropout-probability': model_config['dropout-probability'],
                'receptive-field': model_config['receptive-field'],
                'embedding-size-0': model_config['embedding-size-0'],
                'embedding-size-1': model_config['embedding-size-1'],
                'activation': model_config['activation'],
                'categories': model_config['categories'],
                'model': model.state_dict()
            }, path)
    elif model_config['type'] == 'mbnet0':
        torch.save(
            {
                'type': 'mbnet0',
                'downscale': model_config['downscale'],
                'dropout-probability': model_config['dropout-probability'],
                'embedding-size': model_config['embedding-size'],
                'activation': model_config['activation'],
                'convolution-layers': model_config['convolution-layers'],
                'embedding-expansion-factor':
                model_config['embedding-expansion-factor'],
                'kernel-size': model_config['kernel-size'],
                'categories':  model_config['categories'],
                'model': model.state_dict()
            }, path)


# Focal loss reduces weight of confidently correctly classified samples,
# thereby making training focus on samples where classification is not confident
# or even incorrect.
# https://arxiv.org/abs/1708.02002
# https://discuss.pytorch.org/t/is-this-a-correct-implementation-for-focal-loss-in-pytorch/43327/8
def focal_loss(prediction, target, weight=None, gamma=2., reduction='mean'):
    log_prob = torch.nn.functional.log_softmax(prediction, dim=-1)
    prob = torch.exp(log_prob)
    return torch.nn.functional.nll_loss(((1 - prob)**gamma) * log_prob,
                                        target,
                                        weight=weight,
                                        reduction=reduction)


def evaluate_model(model, batch_input, targets, receptive_field):
    predictions = model(batch_input)[:, receptive_field // 2,
                                     receptive_field // 2, :]
    #loss = torch.nn.functional.cross_entropy(predictions, targets)
    loss = focal_loss(predictions, targets)
    categories = torch.argmax(predictions, dim=1)
    is_correct = categories == targets

    return loss, is_correct


if __name__ == '__main__':
    import argparse
    import json
    import time
    import torchvision

    parser = argparse.ArgumentParser()
    parser.add_argument('--labels',
                        default=None,
                        help='Path to a json file with image annotations.')
    parser.add_argument('--downscale',
                        type=int,
                        default=4,
                        help='How much to shrink the input image.')
    parser.add_argument('--receptive-field',
                        type=int,
                        default=13,
                        help='Width (and height) of an area used for pixel ' +
                        'classification.')
    parser.add_argument('--model-config',
                        type=json.loads,
                        default={
                            "type": "mdnet0",
                            "embedding-size-0": 64,
                            "embedding-size-1": 64,
                            "activation": "relu",
                            "dropout-probability": 0.4
                        })
    parser.add_argument('--batch-size',
                        type=int,
                        default=1024,
                        help='Training batch size.')
    parser.add_argument('--num-loading-workers',
                        type=int,
                        default=4,
                        help='Training batch size.')
    parser.add_argument('--epochs',
                        type=int,
                        default=1000,
                        help='Number of training epochs')
    parser.add_argument('--num-training-transforms',
                        type=int,
                        default=64,
                        help='Number for random transormations of training ' +
                        'patches.')
    parser.add_argument('--num-testing-transforms',
                        type=int,
                        default=1,
                        help='Number for random transormations of testing ' +
                        'patches.')
    parser.add_argument('--a-few-negatives',
                        type=int,
                        default=200,
                        help='Number of negatives taken per image.')
    parser.add_argument('--nothing-stride',
                        type=int,
                        default=10,
                        help='Density of pixels taken from `nothing` ' +
                        'training samples.')
    parser.add_argument('--lr',
                        type=float,
                        default=10.0,
                        help='Learning rate.')
    parser.add_argument('--gamma',
                        type=float,
                        default=0.999,
                        help='Learning rate step gamma.')
    parser.add_argument('--saved-model',
                        default=None,
                        help='Path to where to save the trained model.')
    parser.add_argument('--evaluation-frequency',
                        type=int,
                        default=10,
                        metavar='N',
                        help='Evaluation of the model done every N epochs.')
    args = parser.parse_args()

    use_cuda = torch.cuda.is_available()
    kwargs = {
        'num_workers': args.num_loading_workers,
        'pin_memory': False
    } if use_cuda else {}
    device = torch.device("cuda" if use_cuda else "cpu")
    torch.manual_seed(41)

    image_transforms = torchvision.transforms.Compose([
        torchvision.transforms.RandomAffine(
            degrees=20,
            scale=(0.9, 1.1),
            translate=(1 / args.receptive_field, 1 / args.receptive_field)),
        torchvision.transforms.RandomHorizontalFlip(),
        torchvision.transforms.CenterCrop(args.receptive_field),
        torchvision.transforms.ColorJitter(brightness=(0.8, 1.2),
                                           contrast=(0.8, 1.2),
                                           saturation=(0.8, 1.2),
                                           hue=0.02)
    ])
    training_set = SubtDataset(
        labels_config=args.labels,
        downscale=args.downscale,
        receptive_field=args.receptive_field,
        transform=(image_transforms if args.num_testing_transforms else None),
        num_transforms=args.num_training_transforms,
        a_few_negatives=args.a_few_negatives,
        nothing_stride=args.nothing_stride)
    training_loader = torch.utils.data.DataLoader(training_set,
                                                  batch_size=args.batch_size,
                                                  shuffle=True,
                                                  **kwargs)

    test_set = SubtDataset(labels_config=args.labels,
                           downscale=args.downscale,
                           receptive_field=args.receptive_field,
                           transform=image_transforms,
                           num_transforms=args.num_testing_transforms,
                           a_few_negatives=args.a_few_negatives,
                           nothing_stride=args.nothing_stride)
    test_set = training_set
    test_loader = torch.utils.data.DataLoader(test_set,
                                              batch_size=args.batch_size,
                                              shuffle=False,
                                              **kwargs)

    # No downscaling during training, because dataset loader already takes care
    # of that.
    args.model_config['downscale'] = 1
    args.model_config['receptive-field'] = args.receptive_field
    args.model_config['categories'] = ARTIFACT_CATEGORIES
    model = create_model(args.model_config)

    model = model.float()
    if use_cuda:
        model = model.cuda()

    optimizer = torch.optim.Adadelta(model.parameters(), lr=args.lr)
    scheduler = torch.optim.lr_scheduler.StepLR(optimizer,
                                                step_size=1,
                                                gamma=args.gamma)

    best_accuracy = -1
    for epoch in range(1, args.epochs + 1):
        model.train()
        for batch_idx, (patches, targets,
                        origins) in enumerate(training_loader):
            patches = patches.float().to(device)
            targets = targets.to(device)
            optimizer.zero_grad()
            loss, is_correct = evaluate_model(model, patches, targets,
                                              args.receptive_field)
            accuracy = torch.mean(is_correct.float())
            loss.backward()
            optimizer.step()
            print('TRAIN epoch={}, batch={}, loss={}, accuracy={}'.format(
                epoch, batch_idx, loss.item(), accuracy.item()))

        if epoch % args.evaluation_frequency == 0:
            model.eval()
            n = 0
            total_acc = 0.0
            for batch_idx, (patches, targets,
                            origins) in enumerate(test_loader):
                patches = patches.float().to(device)
                targets = targets.to(device)
                loss, is_correct = evaluate_model(model, patches, targets,
                                                  args.receptive_field)
                total_acc += torch.sum(is_correct.float()).item()
                n += len(patches)
                from_category = dict(
                    (cid, cat) for (cat, cid) in ARTIFACT_CATEGORIES.items())
                for c, o, t in zip(is_correct, origins, targets):
                    if not c.item():
                        print('misclassified:', from_category[t.item()], o)

            acc = total_acc / n
            # Not aggregating and printing `loss`, because the output
            # MaybeSoftMax layer makes it incorrect. CrossEntropyLoss expects
            # logits, not probabilities.
            print('TEST epoch={}, accuracy={}'.format(epoch, acc))
            time.sleep(0.5)  # To gove myself time to spot the message above.
            if acc > best_accuracy:
                best_accuracy = acc
                if args.saved_model:
                    args.model_config['downscale'] = args.downscale
                    save_model(args.model_config, model, args.saved_model)
            if acc == 1.0:
                break

        scheduler.step()



import unittest
from unittest.mock import MagicMock
from datetime import timedelta

from osgar.drivers.eduro import Eduro, sint32_diff, CAN_triplet
from osgar.bus import Bus


class EduroTest(unittest.TestCase):

    def test_sync(self):
        logger = MagicMock()
        logger.write = MagicMock(return_value=timedelta(135))
        bus = Bus(logger)
        eduro = Eduro(config={}, bus=bus.handle('eduro'))
        tester = bus.handle('tester')
        tester.register('can')
        bus.connect('tester.can', 'eduro.can')
        bus.connect('eduro.pose2d', 'tester.pose2d')
        sync = CAN_triplet(0x80, [])
        tester.publish('can', sync)
        eduro.request_stop()
        eduro.run()
        tester.shutdown()
        self.assertEqual(tester.listen(), (timedelta(135), 'pose2d', [0, 0, 0]))

    def test_buttons(self):
        logger = MagicMock()
        logger.write = MagicMock(return_value=timedelta(42))
        bus = Bus(logger)
        eduro = Eduro(config={}, bus=bus.handle('eduro'))
        tester = bus.handle('tester')
        tester.register('can')
        bus.connect('eduro.buttons', 'tester.buttons')
        bus.connect('tester.can', 'eduro.can')
        tester.publish('can', CAN_triplet(0x28A, [0, 0]))
        eduro.request_stop()
        eduro.run()
        tester.shutdown()
        self.assertEqual(tester.listen(), (timedelta(42), 'buttons', {'blue_selected': True, 'cable_in': False}))

    def test_encoders_overflow(self):
        logger = MagicMock()
        logger.write = MagicMock(return_value=timedelta(22))
        bus = Bus(logger)
        eduro = Eduro(config={}, bus=bus.handle('eduro'))
        tester = bus.handle('tester')
        tester.register('can')
        bus.connect('eduro.encoders', 'tester.encoders')
        bus.connect('tester.can', 'eduro.can')

        sync = CAN_triplet(0x80, [])

        enc_left = CAN_triplet(0x181, [0xff, 0xff, 0xff, 0x7f])
        tester.publish('can', enc_left)
        tester.publish('can', sync)

        enc_left = CAN_triplet(0x181, [0x01, 0x00, 0x00, 0x80])
        tester.publish('can', enc_left)
        tester.publish('can', sync)
        eduro.request_stop()
        eduro.run()
        tester.shutdown()
        self.assertEqual(tester.listen(), (timedelta(22), 'encoders', [0, 0]))
        self.assertEqual(tester.listen(), (timedelta(22), 'encoders', [2, 0]))

    def test_sint32_diff(self):
        self.assertEqual(sint32_diff(-5, 7), -12)
        self.assertEqual(sint32_diff(-0x7FFFFFFF, 0x7FFFFFFF), 2)
        self.assertEqual(sint32_diff(0x7FFFFFFF, -0x7FFFFFFF), -2)

    def test_float_speed_bug(self):
        bus = MagicMock()
        eduro = Eduro(config={}, bus=bus)
        # set internal values before crash
        eduro.desired_speed = 0.71
        eduro.desired_angular_speed = 1.6952383024620923
        eduro._rampLastLeft = 1417
        eduro._rampLastRight = 3891
        eduro.send_speed()
        bus.publish.assert_called_once_with('can', [0x201, b'\xa9\x05\xa0\x0f', 0])

# vim: expandtab sw=4 ts=4



import unittest
from unittest.mock import MagicMock

from osgar.drivers.rosmsg import (ROSMsgParser, parse_jpeg_image, parse_laser,
                                  parse_odom, parse_imu, parse_points, get_frame_id,
                                  parse_bool, parse_volatile, parse_bucket, parse_topic, parse_posestamped)


class ROSMsgParserTest(unittest.TestCase):

    def Xtest_parse_imu(self):
        r = ROSMsgParser(config={}, bus=None)
#        with open('imu_data.txt', 'rb') as f:
        with open('imu-x3.bin', 'rb') as f:
            r._buf += f.read()
            index = 0
            packet = r.get_packet()  # first packet is structure file
            while packet is not None:
                if index > 0:
                    ori = parse_imu(packet)
#                    q0, q1, q2, q3 = ori  # quaternion

#                    x =  math.atan2(2*(q0*q1+q2*q3), 1-2*(q1*q1+q2*q2))
#                    y =  math.asin(2*(q0*q2-q3*q1))
#                    z =  math.atan2(2*(q0*q3+q1*q2), 1-2*(q2*q2+q3*q3))
#                    print([math.degrees(a) for a in [x, y, z]])

                packet = r.get_packet()
                index += 1
#                if index > 10000:
#                    break

    def Xtest_parse_image(self):
        r = ROSMsgParser(config={}, bus=None)
        with open('image_raw2.bin', 'rb') as f:
            r._buf += f.read()
            index = 0
            packet = r.get_packet()  # first packet is structure file
            while packet is not None:
                if index > 0:
                    parse_jpeg_image(packet, 'dump_%03d.jpg' % index)
                packet = r.get_packet()
                index += 1
                if index > 10:
                    break

    def Xtest_parse_laser(self):
        r = ROSMsgParser(config={}, bus=None)
        with open('laser_data2.bin', 'rb') as f:
            r._buf += f.read()
            index = 0
            packet = r.get_packet()  # first packet is structure file
            while packet is not None:
                if index > 0:
                    scan = parse_laser(packet)
                    print(scan)
                    self.assertEqual(len(scan), 720)
                    self.assertEqual(type(scan[0]), int)
                packet = r.get_packet()
                index += 1
                if index > 10:
                    break

    def Xtest_parse_odom(self):
        r = ROSMsgParser(config={}, bus=None)
        with open('odom_data.bin', 'rb') as f:
#        with open('x3-odom.bin', 'rb') as f:
            r._buf += f.read()
            index = 0
            packet = r.get_packet()  # first packet is structure file
            while packet is not None:
                if index > 0:
                    parse_odom(packet)
                packet = r.get_packet()
                index += 1
                if index > 10:
                    break

    def Xtest_parse_points(self):
        r = ROSMsgParser(config={}, bus=None)
        with open('point_data.bin', 'rb') as f:
            r._buf += f.read()
            index = 0
            packet = r.get_packet()  # first packet is structure file
            while packet is not None:
                if index > 0:
                    parse_points(packet)
                packet = r.get_packet()
                index += 1
                if index > 10:
                    break

    def test_get_frame_id(self):
        data = b'\x08\x00\x00\x00\x02\x00\x00\x00\x00\x06\x81\x14'
        self.assertEqual(get_frame_id(data), b'/clock')
        data = b'\x01\x00\x00\x00\x00'
        self.assertEqual(get_frame_id(data), b'/gas_detected')

    def test_parse_bool(self):
        data = b'\x01\x00\x00\x00\x00'
        self.assertEqual(parse_bool(data), False)

    def test_parse_volatile(self):
        data = b"<\x00\x00\x00\x8c\x00\x00\x009\x00\x00\x00\x00'\xb9)\x17\x00\x00\x00" + \
               b"scout_1/volatile_sensor\x08\x00\x00\x00methanol\n\x00\x00\x00\x01\xfb\xde`?"
        self.assertEqual(parse_volatile(data), ['methanol', 0.8784024119377136, 10])

    def test_parse_bucket(self):
        data = b'\x16\x00\x00\x00\n\x00\x00\x00sulfur_dio\x1b\x00\x00\x00\xd3\xa7\xabA'
        self.assertEqual(parse_bucket(data), ['sulfur_dio', 27, 21.456945419311523])

    def test_parse_bin(self):
        data = b'\x82\x00\x00\x00\x03\x00\x00\x00ice\x06\x00\x00\x00ethene\x07\x00\x00\x00methane\x0b\x00\x00\x00carbon_mono\n\x00\x00\x00carbon_dio\x07\x00\x00\x00ammonia\x0c\x00\x00\x00hydrogen_sul\n\x00\x00\x00sulfur_dio\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
        self.assertEqual(parse_topic('srcp2_msgs/HaulerMsg', data), [['ice', 'ethene', 'methane', 'carbon_mono', 'carbon_dio', 'ammonia', 'hydrogen_sul', 'sulfur_dio'], [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]])

    def test_parse_posestamped(self):
        data = b"K\x00\x00\x00\x16\x01\x00\x00s\x06\x00\x00@O\xa86\x03\x00\x00\x00map\xcb\x80M+\xc9\xcf\xce?\xb8\xb5\x9c\x04\xfb\x98\xbb\xbfp\xbb\x0e\xe8'U\xd1\xbfa`9&\x83F\x84\xbf\xce\xf7O\x9f\x90e\xbe\xbf\xf7,N\xba\xd7\xcf\xb0?\x8b\xc3D\xf5\xd4\xb3\xef?"
        self.assertEqual(parse_posestamped(data), ((0.24071611990348737, -0.10780304777718797, -0.2708225027262676), (-0.00990011654282358, -0.11873725785696296, 0.06567142771216426, 0.9907021322634234)))

    def test_parse_score_qual2(self):
        data = b'\x87\x00\x00\x00\x03\x00\x00\x00ice\x06\x00\x00\x00ethene\x07\x00\x00\x00methane' + \
               b'\x08\x00\x00\x00carbon_mono\n\x00\x00\x00carbon_dio\x07\x00\x00\x00ammonia\x0c\x00\x00\x00' + \
               b'hydrogen_sul\n\x00\x00\x00sulfur_dio\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' + \
               b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' + \
               b'\x00\x00\x00\x00\x00\x00\x00\x00\x00'
        self.assertEqual(parse_topic('srcp2_msgs/Qual2ScoringMsg', data), [0, 0])

    def test_imu_math_domain_error(self):
        data = b'J\x01\x00\x00\x0c\x18\x01\x00\x05\t\x00\x00\x80\xc0\r\x16\x12\x00\x00\x00scout' + \
               b'_1///imu_linkl\x18B2\xb2e\x95?\xb8T\x9b\x8f\xcc\x1c\xe5\xbf\xb5dw\x91\xf7h\xd8?' + \
               b'\xe5\xaaF\x00\xf1\xb9\xe8\xbf|\x14\xaeG\xe1zd?\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00|\x14\xaeG\xe1zd?\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00|\x14\xaeG\xe1zd?\x80\xa6\xa6\xeam\x1b~?x_\xd7\x19\xb6\x02\xc5?\t\xd4\xe3*\xbbx\xb6\xbf|\x14\xaeG\xe1zd?\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00|\x14\xaeG\xe1zd?\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00|\x14\xaeG\xe1zd?\xb8\xec\x82\xe3\xe1[\xd1\xbf\x14K\xb9TQ]\x03\xc0\xdb\xfc\x99\xf1\xb8\x05\x00@|\x14\xaeG\xe1zd?\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00|\x14\xaeG\xe1zd?\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00|\x14\xaeG\xe1zd?'
        parse_imu(data)  # it failed to compute asin from -1.003659325094236

    def test_radio(self):
        data = b'radio X30F60R [0, 100, 30]\n'
        bus = MagicMock()
        r = ROSMsgParser(config={}, bus=bus)
        r.slot_raw(timestamp=None, data=data)
        bus.publish.assert_called_with('radio', [b'X30F60R', b'[0, 100, 30]\n'])

    def test_publish_desired_speed(self):
        data = b'\x08\x00\x00\x00\x02\x00\x00\x00\x00\x06\x81\x14'  # clock as 8 bytes
        bus = MagicMock()
        r = ROSMsgParser(config={}, bus=bus)
        r.slot_raw(timestamp=None, data=data)
        bus.publish.assert_called_with('sim_time_sec', 2)

        clock_data = b'\x08\x00\x00\x00\x03\x00\x00\x00\x00\x00\x00\x00'  # clock 3s
        r.slot_raw(timestamp=None, data=clock_data)
        bus.publish.assert_called_with('sim_time_sec', 3)  # initially desired speed is None -> no cmd

        r.slot_desired_speed(timestamp=None, data=[0, 0])
        r.slot_raw(timestamp=None, data=clock_data)
        # ... asserting that the last call has been made in a particular way
        bus.publish.assert_called_with('cmd', b'cmd_vel 0.000000 0.000000')

        # after update from 3D desired speed only extended cmd_vel_3d should be used
        r.slot_desired_speed_3d(timestamp=None, data=[[1, 2, 3], [4, 5, 6]])
        r.slot_raw(timestamp=None, data=clock_data)
        bus.publish.assert_called_with('cmd', b'cmd_vel_3d 1.000000 2.000000 3.000000 4.000000 5.000000 6.000000')

    def test_publish_undefined_desired_speed(self):
        # for the drone it is important not to send any command until the very first desired speed is received
        pass

# vim: expandtab sw=4 ts=4



"""
  Osgar Config Class
"""
import json
import numbers
import sys
from importlib import import_module

from osgar.drivers import all_drivers


ROBOT_CONTAINER_VER = 2
SUPPORTED_VERSIONS = [ROBOT_CONTAINER_VER]


def get_class_by_name(name):
    if name in all_drivers:
        name = all_drivers[name]
    assert ':' in name, name  # import path and class name expected
    s = name.split(':')
    assert len(s) == 2  # package and class name
    module_name, class_name = s
    m = import_module(module_name)
    return getattr(m, class_name)


class MergeConflictError(Exception):
    pass


def _application2import(application):
    # ModuleSpec: https://www.python.org/dev/peps/pep-0451/
    # __qualname__: https://www.python.org/dev/peps/pep-3155/
    s = sys.modules[application.__module__].__spec__
    return f"{s.name}:{application.__qualname__}"


def config_load(*filenames, application=None):
    ret = {}
    for filename in filenames:
        with open(filename) as f:
            data = json.load(f)
            assert 'version' in data, data
            assert data['version'] in SUPPORTED_VERSIONS, data['version']
            ret = merge_dict(ret, data)
    if application is None:
        return ret
    if not isinstance(application, str):
        application = _application2import(application)
    for module_name, module_config in ret['robot']['modules'].items():
        if module_config['driver'] == 'application':
            module_config['driver'] = application
    return ret


def merge_dict(dict1, dict2):
    ret = dict1.copy()
    for key, val2 in dict2.items():
        try:
            val1 = dict1[key]
            if isinstance(val1, dict) and isinstance(val2, dict):
                ret[key] = merge_dict(val1, val2)
            else:
                for t in numbers.Number, str, list, tuple:
                    if isinstance(val1, t) and isinstance(val2, t):
                        ret[key] = val2
                        break
                else:
                    raise MergeConflictError(str((val1, val2)))
        except KeyError:
            ret[key] = val2
    return ret

# vim: expandtab sw=4 ts=4



import unittest
import datetime
from unittest.mock import MagicMock

from osgar.drivers.maria import RobotMaria, SpeedControl, MAX_PWM


class RobotTankMariaTest(unittest.TestCase):

    def test_parse_packet(self):
        r = RobotMaria({}, MagicMock())
        self.assertEqual((None, b''), r.get_packet(b''))
        self.assertEqual((b'2897 3219', b''), r.get_packet(b'2897 3219\r\n'))

    def test_invalid_packet(self):
        r = RobotMaria({}, MagicMock())
        self.assertEqual((None, b'\xfe'), r.get_packet(b'\xfe'))
        self.assertEqual((b'\xfe51 0 0', b''), r.get_packet(b'\xfe51 0 0\r\n'))

    def test_speed_control(self):
        sc = SpeedControl()
        self.assertEqual(sc.update(0), 0)

        sc.set_desired_ticks(50)
        self.assertEqual(sc.update(0), 10)  # return MIN_PWM
        self.assertEqual(sc.update(0), 12)  # speed up by step
        self.assertEqual(sc.update(50), 12) # keep last pwm
        self.assertEqual(sc.update(70), 10) # slow down by step

        # change polarity
        sc.set_desired_ticks(-50)
        self.assertEqual(sc.update(50), 0)  # wait until it is not moving wrong direction
        self.assertEqual(sc.update(3), -10) # -MIN_PWM
        self.assertEqual(sc.update(-30), -12)  # -(MIN_PWM + step)
        self.assertEqual(sc.update(-51), -12)  # last_pwm (encoders is within tolerance)
        self.assertEqual(sc.update(-66), -10)  # last_pwm - step

        for __ in range(20):
            self.assertGreaterEqual(sc.update(-10), -MAX_PWM)

        sc.set_desired_ticks(50)
        self.assertEqual(sc.update(-70), 0)
        self.assertEqual(sc.update(-2), 10)
        self.assertEqual(sc.update(-2), 12)

# vim: expandtab sw=4 ts=4



import apriltag
import cv2
import numpy

from osgar.logger import LogReader, lookup_stream_names
from osgar.lib.serialize import deserialize

def main():
    import argparse
    from pprint import pprint

    parser = argparse.ArgumentParser(description='')
    parser.add_argument('log', help='filepaths', nargs='+')
    parser.add_argument('--threads', help='how many threads to use', type=int, default=1)
    parser.add_argument('--gui', help='show found tags', default=False, action='store_true')
    parser.add_argument('--margin', help='apriltag decision margin threshold', default=30, type=int)

    args = parser.parse_args()
    detector = apriltag.apriltag('tag16h5', threads=args.threads)

    for filepath in args.log:
        print(filepath)
        streams = lookup_stream_names(filepath)
        processing = []
        for i, stream in enumerate(streams):
            if (stream.startswith("camera") and stream.endswith("raw")) or stream.endswith('.jpg') or stream.endswith('.image'):
                print("processing stream {} => {}".format(i, stream))
                processing.append(i+1)
        if len(processing) == 0:
            print("no jpeg found in streams:")
            pprint(streams)
            continue

        with LogReader(filepath, only_stream_id=processing) as log:
            for dt, channel, data in log:
                try:
                    jpeg = deserialize(data)
                except Exception as e:
                    print(e)
                    continue
                np_jpeg = numpy.frombuffer(jpeg, dtype='u1')
                gray = cv2.imdecode(np_jpeg, cv2.IMREAD_GRAYSCALE)
                found = detector.detect(gray)
                found = [tag for tag in found if tag['margin'] > args.margin and tag['hamming'] == 0]
                if len(found) > 0:
                    ids = list(tag['id'] for tag in found)
                    print(dt, end=' ')
                    if args.gui:
                        img = cv2.imdecode(np_jpeg, cv2.IMREAD_COLOR)
                    for tag in found:
                        rect = tag['lb-rb-rt-lt'].astype('float32')
                        area = cv2.contourArea(rect)
                        print('{:2d}: {{margin: {:3d}, area: {:4d}}}'.format(tag['id'], int(tag['margin']), int(area)), end=' ')
                        center = tuple(tag['center'].astype(int))
                        poly = rect.astype('int32')
                        if args.gui:
                            cv2.circle(img, center, 3, (0, 0 ,255), -1)
                            cv2.polylines(img, [poly], True, (0, 255, 255), 3)
                    print()
                    if args.gui:
                        cv2.imshow('image', img)
                        key = cv2.waitKey(0)
                        if key == 27:
                            return

if __name__ == "__main__":
    main()





import math

TRACE_STEP = 0.5  # meters in 3D


def distance3D(xyz1, xyz2, weights=[1.0, 1.0, 1.0]):
    return math.sqrt(sum([w * (a-b)**2 for a, b, w in zip(xyz1, xyz2, weights)]))


class Trace:
    def __init__(self, step=TRACE_STEP):
        self.trace = []  # unknown start position
        self.step = step

    def update_trace(self, pos_xyz):
        if not self.trace or distance3D(self.trace[-1], pos_xyz) >= self.step:
            self.trace.append(pos_xyz)

    def prune(self, radius=None):
        # use short-cuts and remove all cycles
        if radius is None:
            radius = self.step

        pruned = Trace(step=self.step)
        if self.trace:
            # keep the very first position in pruned version
            pruned.update_trace(self.trace[0])

        open_end = 1
        while open_end < len(self.trace):
            best = open_end
            for i, xyz in enumerate(self.trace[open_end:], start=open_end):
                if distance3D(xyz, pruned.trace[-1]) < radius:
                    best = i
            pruned.update_trace(self.trace[best])
            open_end = best + 1
        self.trace = pruned.trace


    def where_to(self, xyz, max_target_distance, z_weight=0.2):
        # looking for a target point within max_target_distance nearest to the start
        for _ in range(8):
            for target in self.trace:
                if distance3D(target, xyz, [1.0, 1.0, z_weight]) < max_target_distance:
                    return target
            # if the robot deviated too far from the trajectory, we need to look for more distant target points
            max_target_distance *= 1.5
        # robot is crazy far from the trajectory
        assert(False)

    def reverse(self):
        self.trace.reverse()

    def add_line_to(self, xyz):
        if self.trace:
            last = self.trace[-1]
            size = distance3D(last, xyz)
            dx, dy, dz = xyz[0] - last[0], xyz[1] - last[1], xyz[2] - last[2]
            for i in range(1, int(size/self.step)):
                s = self.step * i/size
                self.trace.append((last[0] + s*dx, last[1] + s*dy, last[2] + s*dz))
        self.trace.append(xyz)

    def start_position(self):
        if self.trace:
            return self.trace[0]
        return None



import unittest
from unittest.mock import MagicMock
from datetime import timedelta

from osgar.bus import Bus
from osgar.node import Node


class NodeTest(unittest.TestCase):

    def test_usage(self):
        empty_config = {}
        bus = Bus(logger=MagicMock())
        node = Node(config=empty_config, bus=bus.handle('mynode'))
        node.start()
        node.request_stop()
        node.join()

    def test_update(self):
        empty_config = {}
        bus = Bus(logger=MagicMock())
        node = Node(config=empty_config, bus=bus.handle('mynode'))
        tester = bus.handle('tester')
        tester.register('vel')
        bus.connect('tester.vel', 'mynode.vel')
        dt = tester.publish('vel', 3)
        node.update()
        self.assertEqual(node.time, dt)
        self.assertEqual(node.vel, 3)

        node2 = Node(config=empty_config, bus=bus.handle('mynode2'))
        self.assertNotIn('vel', dir(node2))

# vim: expandtab sw=4 ts=4



"""
  Convert depth images to B&W where White is danger
"""

import cv2
import numpy as np


class DepthParams:
    def __init__(
            self,
            # Focal length.
            fx=462.1,
            # Placement of the camera relative to the center of the robot.
            camera_xyz=(0.23, 0, (0.19 + 0.06256005)),
            # Orientation of the camera relative to the robot. In radians.
            camera_ypr=(0, 0, 0),
            # Image dimensions.
            image_size=(640, 360),
            # Principal point, position of optical axis
            principal_point=(640 / 2. +  0.5, 360 / 2 + 0.5),
            min_x = 0.,
            min_y = 0.,
            max_x = 16.,
            max_y = 16.,
            # Low-height stuff on the ground does not matter. (In meters.)
            min_z=0.08,
            # Stuff above the robot does not matter.
            max_z=1.5,
            # We compare pixels this far away from each other vertically.
            # (In pixels.)
            vertical_pixel_offset=1,
            # How close to the vertical does an obstacle need to be?
            # (In radians.)
            vertical_diff_limit=np.radians(60),
            # How much of a slope do we still consider "normal".
            max_slope=np.radians(-16),
            slope_length_scale_factor = 1.0,
            slope_length_power_factor = 0.0,
            # Despeckling parameters.
            noise_filter_window=(5, 5),
            noise_filter_threshold=20,
            # Lidar field of view in radians.
            lidar_fov=np.radians(270),
            # Distance to which we fully trust the lidar.
            lidar_trusted_zone=0,
            **kwargs):

        self.fx = fx
        self.camera_xyz = np.asarray(camera_xyz)
        self.camera_rot = _rotation_matrix(*camera_ypr)
        self.camw, self.camh = image_size
        self.cam_low = self.camh // 2 + 1
        self.rx, self.ry = principal_point
        self.min_x = min_x
        self.min_y = min_y
        self.max_x = max_x
        self.max_y = max_y
        self.min_z = min_z
        self.max_z = max_z
        self.vertical_pixel_offset = vertical_pixel_offset
        self.vertical_diff_limit = vertical_diff_limit
        self.max_slope = max_slope
        self.slope_length_scale_factor = slope_length_scale_factor
        self.slope_length_power_factor = slope_length_power_factor
        self.noise_filter_window = noise_filter_window
        self.noise_filter_threshold = noise_filter_threshold
        self.lidar_fov = lidar_fov
        self.lidar_trusted_zone = lidar_trusted_zone

        # Pixel coordinates relative to the center of the image, with positive
        # directions to the left and up.
        pxs = self.rx - np.repeat(
                np.arange(self.camw).reshape((1, self.camw)), self.camh, axis=0)
        pys = self.ry - np.repeat(
                np.arange(self.camh).reshape((self.camh, 1)), self.camw, axis=1)
        pzs = np.ones((self.camh, self.camw), dtype=np.float)
        # For each pixel in the image, a vector representing its corresponding
        # direction in the scene with a unit forward axis.
        self.ps = (self.camera_rot @ np.dstack([pzs, pxs / fx, pys / fx]).T.reshape((3, -1))).reshape((3, self.camw, self.camh)).T


# Indices of directions in a matrix with 3D points.
X, Y, Z = 0, 1, 2


def _rotation_matrix(yaw, pitch, roll):
    # private so it will not be used by other modules -> TODO move to math or quaternion?
    cos_roll = np.cos(roll)
    sin_roll = np.sin(roll)
    roll_m = np.array([[1., 0., 0.], [0., cos_roll, -sin_roll],
                       [0., sin_roll, cos_roll]])
    cos_pitch = np.cos(pitch)
    sin_pitch = np.sin(pitch)
    pitch_m = np.array([[cos_pitch, 0, sin_pitch], [0, 1, 0],
                        [-sin_pitch, 0, cos_pitch]])
    cos_yaw = np.cos(yaw)
    sin_yaw = np.sin(yaw)
    yaw_m = np.array([[cos_yaw, -sin_yaw, 0], [sin_yaw, cos_yaw, 0],
                      [0, 0, 1]])

    return roll_m @ pitch_m @ yaw_m


def depth2danger(depth_mm, params):
    # COPY & PASTE - refactoring needed!

    depth = depth_mm * 0.001  # Converting to meters.
    # 3D coordinates of points detected by the depth camera, converted to
    # robot's coordinate system.
    xyz = (params.ps * np.expand_dims(depth, axis=Z) +
           [[params.camera_xyz]])
    # Relative positions of 3D points placed params.vertical_pixel_offset pixels
    # above each other.
    rel_xyz = (xyz[:-params.vertical_pixel_offset,:,:] -
               xyz[params.vertical_pixel_offset:,:,:])
    # Slope of a line connecting two points in the vertical plane they
    # define.
    slope = np.arctan2(
            rel_xyz[:,:,Z], np.hypot(rel_xyz[:,:,X], rel_xyz[:,:,Y]))

    nearer_x = np.minimum(xyz[:-params.vertical_pixel_offset,:,X],
                          xyz[params.vertical_pixel_offset:,:,X])
    abs_y = np.abs(xyz[:,:,Y])
    nearer_abs_y = np.minimum(abs_y[:-params.vertical_pixel_offset,:],
                              abs_y[params.vertical_pixel_offset:,:])
    danger = np.logical_and.reduce(np.stack([
        # It has to be near.
        nearer_x <= params.max_x,
        nearer_abs_y <= params.max_y,
        # But not too near.
        nearer_x >= params.min_x,
        nearer_abs_y >= params.min_y,
        # It should not be something small on the ground.
        np.maximum(xyz[:-params.vertical_pixel_offset,:,Z],
                   xyz[params.vertical_pixel_offset:,:,Z]) >= params.min_z,
        # It has to be close to vertical.
        np.abs(slope - np.radians(90)) <= params.vertical_diff_limit,
        # It should not be too high above ground.
        np.maximum(xyz[:-params.vertical_pixel_offset, :, Z],
                   xyz[params.vertical_pixel_offset:, :, Z]) <= params.max_z]))

    # Filter out noise.
    danger = cv2.filter2D(
            danger.astype(np.float),
            -1,
            np.ones(params.noise_filter_window)) > params.noise_filter_threshold

    return danger


def depth2dist(depth_mm, params, pitch=None, roll=None, debug_col=None):
    # return line in mm corresponding to scan
    # optional pitch and roll angles are in radians
    # warning: there is a bug (?) in either Osgar or in SubT and the angles
    # are as if the robot was upside down.
    pitch_rot = np.asmatrix(np.eye(3))
    roll_rot = np.asmatrix(np.eye(3))
    # http://planning.cs.uiuc.edu/node102.html
    if pitch is not None:
        inv_pitch = -pitch # Pitch corrected for the upside down bug.
        pitch_rot = np.matrix(
                [[np.cos(inv_pitch), 0.0, np.sin(inv_pitch)],
                 [0.0, 1.0, 0.0],
                 [-np.sin(inv_pitch), 0.0, np.cos(inv_pitch)]])
    if roll is not None:
        inv_roll = -(roll + np.pi) # Roll corrected for the upside down bug.
        roll_rot = np.matrix(
                [[1.0, 0.0, 0.0],
                 [0.0, np.cos(inv_roll), -np.sin(inv_roll)],
                 [0.0, np.sin(inv_roll), np.cos(inv_roll)]])
    rot = pitch_rot * roll_rot

    depth = depth_mm * 0.001  # Converting to meters.
    # 3D coordinates of points detected by the depth camera, converted to
    # robot's coordinate system.
    rxyz = (params.ps * np.expand_dims(depth, axis=Z) +
            params.camera_xyz.reshape((1, 1, 3)))
    # Compensate for robot's pitch & roll.
    gxyz = np.asarray(
        (rot * np.asmatrix(rxyz.reshape((-1, 3)).T)).T).reshape(
                (params.camh, params.camw, 3))
    # Relative positions of 3D points placed params.vertical_pixel_offset pixels
    # above each other.
    rel_xyz = (gxyz[:-params.vertical_pixel_offset,:,:] -
               gxyz[params.vertical_pixel_offset:,:,:])
    # Slope of a line connecting two points in the vertical plane they
    # define.
    slope = np.arctan2(
            rel_xyz[:,:,Z], np.hypot(rel_xyz[:,:,X], rel_xyz[:,:,Y]))

    abs_y = np.abs(rxyz[:,:,Y])
    sensing_area = np.logical_and.reduce(np.stack([
        # It has to be near.
        rxyz[:-params.vertical_pixel_offset,:,X] <= params.max_x,
        rxyz[params.vertical_pixel_offset:,:,X] <= params.max_x,
        abs_y[:-params.vertical_pixel_offset,:] <= params.max_y,
        abs_y[params.vertical_pixel_offset:,:] <= params.max_y,
        # But not too near.
        rxyz[:-params.vertical_pixel_offset,:,X] >= params.min_x,
        rxyz[params.vertical_pixel_offset:,:,X] >= params.min_x,
        abs_y[:-params.vertical_pixel_offset,:] >= params.min_y,
        abs_y[params.vertical_pixel_offset:,:] >= params.min_y]))

    # Surfaces that are vertical enough to be considered not traversable.
    vertical_obstacle = np.logical_and.reduce(np.stack([
        sensing_area,
        # It should not be something small on the ground.
        np.maximum(rxyz[:-params.vertical_pixel_offset,:,Z],
                   rxyz[params.vertical_pixel_offset:,:,Z]) >= params.min_z,
        # It has to be close to vertical.
        np.abs(slope - np.radians(90)) <= params.vertical_diff_limit,
        # It should not be too high above ground.
        np.maximum(rxyz[:-params.vertical_pixel_offset, :, Z],
                   rxyz[params.vertical_pixel_offset:, :, Z]) <= params.max_z]))

    # Slopes
    rel_dists = np.hypot(rel_xyz[:,:,X], rel_xyz[:,:,Y])
    slopes = np.arctan2(rel_xyz[:,:,Z], rel_dists)
    # This threshold parametrization, with positive scale and negative power
    # factor, allows us to accept short/low steep slopes as traversable, yet
    # mark long steps with smaller slope as dangerous.
    # Short & low "steps" (=steep slopes)  arise on normal surface bumpiness,
    # while long steps show up when the robot approaches a cliff with ground at
    # a lower level below it.
    slope_thresholds = params.max_slope * ((params.slope_length_scale_factor * rel_dists) ** params.slope_length_power_factor)
    steep = np.logical_and(sensing_area, slopes < slope_thresholds)
    if debug_col is not None:
        for info in enumerate(
                zip(np.degrees(slopes[:,debug_col]),
                    rel_dists[:,debug_col],
                    np.degrees(slope_thresholds[:,debug_col]),
                    rel_xyz[:,debug_col,:].tolist(),
                    np.hypot(gxyz[params.vertical_pixel_offset:,debug_col,X],
                    gxyz[params.vertical_pixel_offset:,debug_col,Y]),
                    depth[params.vertical_pixel_offset:,debug_col],
                    steep[:,debug_col])):
            print(info)
        print(np.any(steep[:,debug_col]), steep.shape[0] - np.argmax(steep[::-1,debug_col]) - 1)

    noisy_danger = np.logical_or(vertical_obstacle, steep)

    # Filter out noise.
    danger = cv2.filter2D(
            noisy_danger.astype(np.float),
            -1,
            np.ones(params.noise_filter_window)) > params.noise_filter_threshold

    dists = np.hypot(rxyz[:,:,X], rxyz[:,:,Y])
    FAR_AWAY = 1000.0
    dists = np.where(danger, dists[params.vertical_pixel_offset:,:], FAR_AWAY)
    scan = np.min(dists, axis=0)
    mask = scan == FAR_AWAY
    scan[mask] = 0

    scan = np.array(scan*1000, dtype=np.int32)
    return scan


if __name__ == '__main__':
    import argparse
    import cv2
    import json
    import matplotlib.pyplot as plt
    from osgar.lib.quaternion import euler_zyx

    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument('filename', help='NPZ file with depth data')
    parser.add_argument('--depth-params', type=json.loads, default={},
                        help='Depth processing parameters in json format. ' +
                              'See DepthParams.')
    parser.add_argument('--debug-column', type=int,
                        help='Print extra debugging information for this column in the image.')
    args = parser.parse_args()

    with np.load(args.filename, allow_pickle=True) as f:
        depth = f['depth']
        try:
            xyz, rot = f['pose3d']
            yaw, pitch, roll = euler_zyx(rot)

            # Mimicking Osgar's upside down bug.
            pitch = -pitch
            roll = -roll - np.pi
        except:
            assert(False)
            pitch = 0.0
            roll = 0.0

        try:
            img = f['img']
            img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        except:
            img = None

    depth_params = DepthParams(**args.depth_params)
    dist = depth2dist(depth, depth_params, pitch, roll, debug_col=args.debug_column)

    plt.figure(1)
    plt.plot(range(640), dist, 'o-', linewidth=2)

    depth_img = (np.minimum(255*40, depth) / 40).astype(np.uint8)
    depth_im_color = cv2.applyColorMap(depth_img, cv2.COLORMAP_JET)
    plt.figure(2)
    plt.imshow(depth_im_color)

    if img is not None:
        plt.figure(3)
        plt.imshow(img)

    plt.show()

# vim: expandtab sw=4 ts=4



import collections
import logging
import math
import osgar.node

import osgar.lib.quaternion as quaternion

g_logger = logging.getLogger(__name__)

Pose2d = collections.namedtuple("Pose2d", ("x", "y", "heading"))

class Localization(osgar.node.Node):
    def __init__(self, config, bus):
        super().__init__(config, bus)
        # outputs
        bus.register('pose3d')
        # inputs: origin, orientation, odom
        self.xyz = None # 3D position updated using odometry dist and imu orientation, defined when 'origin' received
        self.orientation = None # not defined until first 'orientation' received
        self.last_odom = None
        self.origin_xyz = None
        self.origin_orientation = None
        self.origin_error = False

    def on_origin(self, data):
        if len(data) == 8:
            self.xyz = data[1:4]
            self.origin_xyz = data[1:4]
            self.origin_orientation = data[4:]
        else:
            self.origin_error = True

    def on_orientation(self, orientation):
        self.orientation = orientation

    def on_odom(self, pose2d):
        x, y, heading = pose2d
        assert self.xyz is not None
        if self.orientation is None:
            return

        odom = Pose2d(x / 1000.0, y / 1000.0, math.radians(heading / 100.0))
        if self.last_odom is not None:
            dist = math.hypot(odom.x - self.last_odom.x, odom.y - self.last_odom.y)
            direction = ((odom.x - self.last_odom.x) * math.cos(self.last_odom.heading) +
                         (odom.y - self.last_odom.y) * math.sin(self.last_odom.heading))
            if direction < 0:
                dist = -dist
        else:
            dist = 0.0

        self.last_odom = odom
        dist3d = quaternion.rotate_vector([dist, 0, 0], self.orientation)
        self.xyz = [a + b for a, b in zip(self.xyz, dist3d)]
        self.bus.publish('pose3d', [self.xyz, self.orientation])

    def run(self):
        try:
            # wait for valid origin
            while True:
                dt, channel, data = self.bus.listen()
                if channel == "origin":
                    self.on_origin(data)
                    if self.xyz is not None:
                        break

            while True:
                dt, channel, data = self.bus.listen()
                self.time = dt
                if channel == "orientation":
                    self.on_orientation(data)
                elif channel == "odom":
                    self.on_odom(data)
                elif channel == "origin":
                    pass
                else:
                    assert False, "unknown input channel"
        except osgar.bus.BusShutdownException:
            pass






import cv2
import functools
import logging
import numpy as np
import sys

import osgar.bus
import osgar.node
import osgar.lib.quaternion

g_logger = logging.getLogger(sys.modules[__name__].__spec__.name)

def no_change_rigid_transform():
    """Create a 4x4 matrix representing no change in orientation or position."""
    return np.eye(4)


def as_rigid_transform(xyz=None, rpy=None, rot=None):
    """Create a rigid transformation matrix representing motion.

    xyz -- X, y and z translation.
    rpy -- Roll, pitch and yaw.
    rot -- 3x3 rotation matrix.

    Only one of `rot` and `rpy` can be set.
    """
    assert (rpy is None or rot is None)
    rt = no_change_rigid_transform()
    if xyz is not None:
        rt[:3, 3] = np.asarray(xyz)
    if rpy is not None:
        rt[:3, :3] = euler_angles_to_rotation_matrix(*rpy)
    if rot is not None:
        rt[:3, :3] = np.asarray(rot)
    return rt


def inverse_rotation(r):
    """Invert a 3x3 rotation matrix.

    https://en.wikipedia.org/wiki/Rotation_matrix#Multiplication
    """
    assert (r.shape == (3, 3))
    return r.T


def inverse_rigid_transform(transform):
    """Invert a 4x4 rigid transformation matrix.

    A rigid transformation matrix consists of a 3x3 rotation submatrix (top
    left) 3x1 translation submatrix and a padding 'to make it work':
        |R t|
        |0 1]

    http://negativeprobability.blogspot.com/2011/11/affine-transformations-and-their.html
    """
    transform = np.asarray(transform)
    assert (transform.shape == (4, 4))
    rot = transform[:3, :3]
    rot_inv = inverse_rotation(rot)
    translation = transform[:3, 3]
    translation_inv = -rot_inv @ translation
    inverse = as_rigid_transform(xyz=translation_inv, rot=rot_inv)
    return inverse


def rotation_matrix_to_euler_angles(rot):
    """Invert a 3x3 rotation matrix to roll, pitch and yaw.

    https://stackoverflow.com/questions/43364900/rotation-matrix-to-euler-angles-with-opencv
    """
    rot = np.asarray(rot)
    assert (rot.shape == (3, 3))
    cosine_for_pitch = np.sqrt(rot[0][0]**2 + rot[1][0]**2)
    is_singular = cosine_for_pitch < 10**-6
    if not is_singular:
        yaw = np.arctan2(rot[1][0], rot[0][0])
        pitch = np.arctan2(-rot[2][0], cosine_for_pitch)
        roll = np.arctan2(rot[2][1], rot[2][2])
    else:
        yaw = np.arctan2(-rot[1][2], rot[1][1])
        pitch = np.arctan2(-rot[2][0], cosine_for_pitch)
        roll = 0
    return roll, pitch, yaw


def euler_angles_to_rotation_matrix(roll, pitch, yaw):
    """Convert roll, pitch and yaw into a 3x3 rotation matrix.

    http://planning.cs.uiuc.edu/node102.html
    """
    cos_roll = np.cos(roll)
    sin_roll = np.sin(roll)
    roll_m = np.array([[1., 0., 0.], [0., cos_roll, -sin_roll],
                       [0., sin_roll, cos_roll]])
    cos_pitch = np.cos(pitch)
    sin_pitch = np.sin(pitch)
    pitch_m = np.array([[cos_pitch, 0, sin_pitch], [0, 1, 0],
                        [-sin_pitch, 0, cos_pitch]])
    cos_yaw = np.cos(yaw)
    sin_yaw = np.sin(yaw)
    yaw_m = np.array([[cos_yaw, -sin_yaw, 0], [sin_yaw, cos_yaw, 0],
                      [0, 0, 1]])

    return yaw_m @ pitch_m @ roll_m


def near_rotation_matrix(m):
    """Find a rotation matrix as close as possible to the input matrix.

    For some definition of "as close as".
    http://users.cecs.anu.edu.au/~hongdong/rotationaveraging.pdf

    This is useful to get rid of errors accumulated in a rotation matrix
    by repeatedly applying matrix multiplication.
    """
    u, s, v = np.linalg.svd(m)
    return u @ v


class VisualOdometryParams:
    def __init__(self,
                 camera_focal_length,
                 camera_center,
                 min_valid_depth=0.01,
                 max_valid_depth=10.,
                 max_corners=1000,
                 corner_quality_level=0.005,
                 corner_distance=5,
                 subpix_win=1,
                 subpix_iters=8,
                 corner_matching_window_size=25,
                 ransac_reprojection_error=0.4,
                 ransac_confidence=0.9999,
                 ransac_pnp_method=cv2.SOLVEPNP_UPNP,
                 ransac_attempts=1,
                 ransac_attempt_reprojection_increase_factor=2.,
                 min_inliers=128,
                 max_rotation_since_key_frame=np.radians(60),
                 max_translation_since_key_frame=2.,
                 max_avg_static_scene_intensity_change=5):
        """Encapsulate parameters needed for visual odometry.

        camera_focal_length -- Focal length of the camera, in pixels.
                               Assuming fx == fy.
        camera_center -- Principal point of the camera, in pixels.
        min_valid_depth -- Minimum depth still considered to be valid, in
                           meters.
        max_valid_depth -- Maximum depth still considered to be valid, in
                           meters.
        max_corners -- Maximum number of feature points detected in the image.
                       Higher number leads to slower and more precise odometry.
        corner_quality_level -- Quality threshold for detected feature points.
        corner_distance -- Minimum required distance between two feature points.
        subpix_win -- Window size for subpixel feature point search.
        subpix_iters -- Number of iterations when searching for subpixel feature
                        points.
        corner_matching_window_size -- Size of the window when matching feature
                                       points between different images.
        ransac_reprojection_error -- Maximum acceptable projection error when
                                     estimating camera movement. Higher value
                                     leads to higher number of less precise
                                     inliers. It is better to keep the value low
                                     while high enough to get sufficiently many
                                     intliers.
        ransac_confidence -- Probability that the estimation of movement
                             produced a useful result.
        ransac_pnp_method -- Which PnP method to use. SOLVE_UPNP and SOLVE_EPNP
                             appear to work well, SOLVE_ITERATIVE doesn't.
        ransac_attempts -- How many movement estimates to perform when there
                           are not enough inliers while gradually increasing
                           acceptable reprojection error.
        ransac_attempt_reprojection_increase_factor -- Multiplicative factor
             gradually increasing acceptable reprojection error while looking
             for sufficiently many inliers.
        min_inliers -- Minimum number of inliers for a movement estimate to be
                       considered valid. Lower number of inliers leads to less
                       frequent creation of key frames and thus to lower error
                       accumulation over key frames. However, it also leads to
                       less precise estimates of the movements since and between
                       key frames. That in turn leads to higher total error. The
                       right value of this parameter needs to hit the sweet spot
                       in this trade-off.
        max_rotation_since_key_frame -- Maximum rotation since key frame that
             we consider OK to still trust the estimate. In radians.
        max_translation_since_key_frame -- Maximum translation since key frame
            that we consider OK to still trust the estimate. In meters.
        max_avg_static_scene_intensity_change
        """
        self.camera_matrix = np.zeros((3, 3))
        self.camera_matrix[0, 0] = camera_focal_length
        self.camera_matrix[1, 1] = camera_focal_length
        self.camera_matrix[0, 2] = camera_center[0]
        self.camera_matrix[1, 2] = camera_center[1]

        self.min_valid_depth = min_valid_depth
        self.max_valid_depth = max_valid_depth

        self.max_corners = max_corners
        self.corner_quality_level = corner_quality_level
        self.corner_distance = corner_distance

        self.subpix_win = subpix_win
        self.subpix_iters = subpix_iters

        self.corner_matching_window_size = corner_matching_window_size
        self.ransac_reprojection_error = ransac_reprojection_error
        self.ransac_confidence = ransac_confidence
        self.ransac_pnp_method = ransac_pnp_method
        self.ransac_attempts = ransac_attempts
        self.ransac_attempt_reprojection_increase_factor = (
            ransac_attempt_reprojection_increase_factor)
        self.min_inliers = min_inliers
        self.max_rotation_since_key_frame = max_rotation_since_key_frame
        self.max_translation_since_key_frame = max_translation_since_key_frame
        self.max_avg_static_scene_intensity_change = (
                max_avg_static_scene_intensity_change)


class CameraFrame:
    """Encapsulates and pre-processes image and depth data."""
    def __init__(self, img, depth, params):
        """Initialize CameraFrame.

        img -- Greyscale image.
        depth -- np.float32 depth image in meters.
        params -- VisualOdometryParams.
        """
        self.img = img.copy()
        self.depth = depth.copy()
        self.depth_is_valid = np.logical_and(
            self.depth >= params.min_valid_depth,
            self.depth <= params.max_valid_depth)

        # In the Virtual Track of Subterranean Challenge, the world is not rich
        # enough for more advanced features, such as SURF, to work well. Corner
        # detectors seem to be faster and work better.
        self.corners = cv2.goodFeaturesToTrack(
                self.img, params.max_corners, params.corner_quality_level,
                params.corner_distance, None,
                self.depth_is_valid.astype(np.uint8))

        if self.corners is not None:
            self.corners = self.corners[:,0,:]

        if (self.corners is not None and self.corners.shape[0] > 0
                and params.subpix_iters > 0):
            term_crit = (cv2.TermCriteria_COUNT, params.subpix_iters, -1)
            self.corners = cv2.cornerSubPix(
                    img, self.corners, (params.subpix_win, params.subpix_win),
                    (-1, -1), term_crit)

            # Subpixel corner detection sometimes yields points outside of the
            # input image, which causes troubles with their processing. Avoid
            # them.
            in_picture = np.logical_and.reduce([
                self.corners[:, 0] >= 0, self.corners[:, 0] < img.shape[1],
                self.corners[:, 1] >= 0, self.corners[:, 1] <= img.shape[0]
            ])
            self.corners = self.corners[in_picture]
        else:
            self.corners = np.zeros((0, 2), dtype=np.float32)


# Conversion transform from OpenCV's coordinates to Osgar's.
# In Osgar, X points forward, Y left and Z up.
# In OpenCV, X points right, Y down and Z forward.
CV2OSGAR = np.array([[0, 0, 1, 0],
                     [-1, 0, 0, 0],
                     [0, -1, 0, 0],
                     [0, 0, 0, 1]])
# Conversion transform from Osgar's coordinate frame to OpenCV's.
OSGAR2CV = inverse_rigid_transform(CV2OSGAR)


def osgar_xyz_to_opencv(pts):
    """Transform (X, Y, Z, 1) points from OpenCV's coordinate system to Osgar's.

    pts -- Row-major list of point coordinates.
    """
    pts = np.asarray(pts)
    assert(pts.shape[-1] == 4)
    return (OSGAR2CV @ pts.T).T


class KeyFrame:
    """Represents a preprocessed key frame for visual odometry.

    Preprocessing consists of computing XYZ coordinates of feature points 
    observed in the frame (= image + depth) both in Osgar's coordinate system
    (X forward, Y left, Z up) and in OpenCV's (X right, Y down, Z forward).

    Useful object properties: `corners_xyz` and `corners_cv_xyz`.
    """
    def __init__(self, frame, params):
        """Initialize and preprocess the key frame.

        frame -- CameraFrame with image & depth information and detected feature
                 points.
        params -- VisualOdometryParams.
        """
        self.frame = frame

        assert (frame.corners is not None)
        # For best results, we should do a bilinear (?) interpolation between
        # depths in the surrounding four pixels around the subpixel feature
        # coordinate. The current solution is simpler and seems to work well
        # enough.
        corner_idxs = np.rint(self.frame.corners).astype(np.int)
        corner_depths = frame.depth[corner_idxs[:, 1], corner_idxs[:, 0]]
        camera_f = params.camera_matrix[0, 0]
        assert(camera_f == params.camera_matrix[1, 1])
        assert (camera_f > 0)
        camera_cx = params.camera_matrix[0, 2]
        camera_cy = params.camera_matrix[1, 2]
        # Coordinates in Osgar's system, i.e. "x forward".
        self.corners_xyz = (np.vstack([
            np.ones((corner_idxs.shape[0], )),
            (camera_cx - corner_idxs[:, 0]) / camera_f,
            (camera_cy - corner_idxs[:, 1]) / camera_f,
        ]).T * corner_depths.reshape((-1, 1))).astype(np.float32)
        if self.corners_xyz.shape[0] > 0:
            self.corners_xyz = cv2.convertPointsToHomogeneous(
                    self.corners_xyz)[:,0,:]
        else:
            # Something to keep us from crashing later when there is nothing
            # to work with.
            self.corners_xyz = np.zeros((0, 4), dtype=np.float32)

        # To OpenCV's coordinate system, i.e. "z forward".
        self.corners_cv_xyz = osgar_xyz_to_opencv(self.corners_xyz)


class CameraVisualOdometry:
    """RGBD visual odometry in a camera-centric coordinate system.

    It reports pose relative to the starting position of the camera. Most
    likely, you should use `RobotVisualOdometry` or `GlobalVisualOdometry`
    wrappers instead of using `CameraVisualOdometry` directly.

    The odometry algorithm utilizes information from a depth camera. Optionally,
    it can incorporate trusted information about heading, for example from an
    IMU + accelerometer + compass sensor, and from an altimeter.

    Terminology: Frame = image + depth

    The algorithm starts with creating a key frame. The following frames are
    localized relative to the key frame by minimizing projection errors of
    feature points from the key frame in the new image. If the algorithm fails
    to find satisfactory position estimate of the new frame relative to the key
    frame, it replaces the key frame with a new key frame created from the last
    preceding frame (i.e. a frame closer to the current frame) and tries again.
    With such new key frame in place, the cycle repeats.

    The algorithm reverts to a constant motion assumption if all attempts to
    estimate what is happening fail.

    The optional supplemental information about orientation and altitude are
    fully trusted and replace corresponding estimates from odometry.

    This is pure odometry and does not close loops and does not try to
    re-localize itself after getting lost.

    There is no guarantee on smoothness of the produced estimates.
    """
    def __init__(self, odometry_params):
        """Initialize CameraVisualOdometry

        odometry_params -- VisualOdometryParams.
        """
        self.params = odometry_params

        self.prev_frame = None
        self.key_frame = None
        self.key_frame_pose = None

        self.trajectory = [no_change_rigid_transform()]

    def can_advance_keyframe(self):
        """Return `True` iff `self.key_frame` can be updated to a new frame."""
        return (self.key_frame is not None
                and self.key_frame.frame is not self.prev_frame)

    def advance_keyframe(self):
        """Turn the last observed frame into a new key frame."""
        self.key_frame = KeyFrame(self.prev_frame, self.params)
        self.key_frame_pose = self.pose

    @property
    def xyz(self):
        "XYZ coordinates relative to initial position of the camera in meters."
        return self.pose[:3, 3]

    @property
    def rpy(self):
        """Roll, pitch and yaw relative to initial orientation of the camera.

        In radians.
        """
        return rotation_matrix_to_euler_angles(self.pose[:3, :3])

    @property
    def pose(self):
        """ Rotation & translation matrix representing current pose.

        4x4 rigid transformation matrix capturing current rotation and
        translation of the camera relative to its initial position.

        I.e., this matrix transforms from coordinate system centered at the
        *current* camera location to a coordinate frame centered at the initial
        camera location.

        |R t|
        |0 1|
        """
        return self.trajectory[-1]

    def update(self, img, depth, camera_orientation=None, altitude=None,
               visualize=False):
        """Update the odometry estimate, extend `self.trajectory`.

        Depending on circumstances, key frame may get advanced.

        img -- Greyscale image.
        depth -- np.float32 depth map with the same resolution as `img`.
        camera_orientation -- Optional 3x3 rotation matrix representing
                              orientation *of the camera* that we want to get
                              *after* the update finishes.
        altitude -- Optional altitude *of the camera* that we want to get
                    *after* the update finishes.
        visualize -- If set to `True`, display an extra visualization with
                     detected image features and their tracking.
        """
        # Warning for developers:
        #  This function can recursively call itself. Be careful about infinite
        #  recursion. The recursive call happens when the estimation fails, but
        #  the used key frame is old-ish and can be replaced with the previous
        #  frame. After replacing the old-ish key frame, the function calls
        #  itself in what is essentially "GOTO Line 0 & try again". The idea is
        #  that should happen at most once per input image.
        assert(len(img.shape) == 2)  # Expecting a greyscale image.
        frame = CameraFrame(img, depth, self.params)


        camera_f = self.params.camera_matrix[0, 0]
        camera_cx = self.params.camera_matrix[0, 2]
        camera_cy = self.params.camera_matrix[1, 2]

        if self.key_frame is None:
            # We have just started.
            self.prev_frame = frame
            self.advance_keyframe()
            return self.pose

        # Special case for static scenes, when the robot is not moving.
        if (np.mean(np.abs(self.prev_frame.img.astype(np.int) -
                           img.astype(np.int))) <=
            self.params.max_avg_static_scene_intensity_change):
            self.prev_frame = frame
            self.trajectory.append(self.pose.copy())
            return self.pose

        # Assuming constant motion and constant time between two consecutive
        # frames, guess that the current movement is similar to the previous
        # one.
        # `rt` stands for "rotation & translation".
        if len(self.trajectory) >= 2:
            prev_rt = (inverse_rigid_transform(self.trajectory[-2]) @
                       self.trajectory[-1])
        else:
            # No movement yet.
            prev_rt = no_change_rigid_transform()

        # Reading from the right:
        #  * Convert from points in coordinate system centered at key frame to
        #    a coordinate system centered where the camera started.
        #  * Convert back from that coordinate system to a coordinate system
        #    centered at our current guessed pose, which is based on previous
        #    pose extended with the same movement as observed in previous
        #    update (= assumption of constant velocity and rotation).
        guessed_pose = self.pose @ prev_rt
        initial_estimate = (inverse_rigid_transform(guessed_pose) @
                            self.key_frame_pose)
        assert (tuple(initial_estimate.shape) == (4, 4))
        # This is where we expect feature points in the key frame to be relative
        # to our guessed current position.
        curr_corners_xyz = np.asarray(
            (initial_estimate @ self.key_frame.corners_xyz.T).T)

        # To OpenCV's coordinate system.
        curr_corners_cv_xyz = osgar_xyz_to_opencv(curr_corners_xyz)

        # Initial guess of pixel coordinates of key frame feature points
        # in the current image.
        if curr_corners_cv_xyz.shape[0] > 0:
            dist_coeffs = None
            curr_corners_guess = cv2.projectPoints(
                # Discarding the 1.0 last column from homogeneous
                # coordinates.
                curr_corners_cv_xyz[:, :3],
                np.zeros(3,),  # rvec
                np.zeros(3,),  # tvec
                self.params.camera_matrix,
                dist_coeffs)[0].reshape((-1, 2)).astype(np.float32)
        else:
            # Something not None to avoid crashes later.
            curr_corners_guess = np.zeros((0, 2), dtype=np.float32)

        # Ideally, we would precompute the pyramids and re-use them,
        # but no luck: "Can't pass precalculated pyramids to
        #              calcOpticalFlowPyrLK in Python"
        #              https://github.com/opencv/opencv/issues/4777
        curr_corners, status, _ = cv2.calcOpticalFlowPyrLK(
            self.key_frame.frame.img,
            frame.img,
            self.key_frame.frame.corners,
            # This copy is, strictly speaking not needed, but avoiding
            # OpenCV overwriting existing data makes debugging easier,
            # because it allows us to compare our initial guess with
            # the output of PyrLK detection.
            curr_corners_guess.copy(),
            winSize=(self.params.corner_matching_window_size,
                     self.params.corner_matching_window_size),
            flags=cv2.OPTFLOW_USE_INITIAL_FLOW)
        if curr_corners is None:
            # To keep the code below happy.
            curr_corners = np.zeros((0, 2), dtype=np.float32)
            status = np.zeros((0, 1), dtype=np.int)

        status = status.astype(np.bool).reshape((-1, ))

        # Only keep corners we could track from key frame to the current image.
        kf_corners = self.key_frame.frame.corners[status]
        curr_corners = curr_corners[status]
        kf_corners_xyz = self.key_frame.corners_xyz[status]
        kf_corners_cv_xyz = self.key_frame.corners_cv_xyz[status]

        # If not enough tracked points and key frame is old-ish,
        # turn the previous frame into a key frame and try an update
        # with that.
        if (kf_corners.shape[0] < self.params.min_inliers
                and self.can_advance_keyframe()):
            self.advance_keyframe()
            return self.update(img, depth, camera_orientation, altitude,
                               visualize)

        # Given feature points in key frame, their 3D coordinates relative to
        # the key frame pose and their projection in the current image, estimate
        # the movement between the key frame and current pose.
        ret = False
        inliers = []
        if kf_corners_cv_xyz.shape[0] >= self.params.min_inliers:
            try:
                # `guess` is the same thing as `initial_estimate`, i.e.
                # our estimate of current movement based on previous movement,
                # except it is in OpenCV's coordinate system style (z forward)
                # instead of in Osgar-style system (x forward).
                guess = OSGAR2CV @ initial_estimate @ CV2OSGAR
                guessed_rvec = np.array(cv2.Rodrigues(guess[:3, :3])[0],
                                        dtype=np.float32)
                guessed_tvec = np.array(guess[:3, 3], dtype=np.float32)
                max_reprojection_error = self.params.ransac_reprojection_error
                dist_coeffs = None
                for _ in range(self.params.ransac_attempts):
                    ret, rvec, tvec, inliers = cv2.solvePnPRansac(
                        kf_corners_cv_xyz[:, :3],
                        curr_corners.astype(np.float32),
                        self.params.camera_matrix,
                        dist_coeffs,
                        # Copy not needed, except to make debugging easier.
                        guessed_rvec.copy(),
                        guessed_tvec.copy(),
                        # According to some, but not all, documentation,
                        # only SOLVE_PNP_ITERATIVE uses the initial
                        # estimate. But that one, empirically speaking, does
                        # not work well.
                        # With OpenCV 4.2, enabling usage of the extrinsic guess
                        # indeed does to change the results. At least with
                        # SOLVEPNP_UPNP.
                        useExtrinsicGuess=True,
                        reprojectionError=max_reprojection_error,
                        confidence=self.params.ransac_confidence,
                        flags=self.params.ransac_pnp_method)
                    if ret:
                        break
                    max_reprojection_error *= (
                        self.params.ransac_attempt_reprojection_increase_factor
                    )
            except cv2.error as e:
                g_logger.error('RANSAC is unhappy: {e}')
            if inliers is None:
                # Like in the song by Nazareth: "None hurts ...". Or was it
                # love?
                inliers = []

        # If we don't like the solution and can try with a fresher key frame,
        # we try again with a fresher key frame.
        if ((not ret or len(inliers) < self.params.min_inliers)
                and self.can_advance_keyframe()):
            self.advance_keyframe()
            return self.update(img, depth, camera_orientation, altitude,
                               visualize)

        if ret:  # OK, so we got a solution ...
            transform = as_rigid_transform(xyz=tvec[:, 0],
                                           rot=cv2.Rodrigues(rvec)[0])
            # We need the transform in Osgar's coordinate system.
            transform = CV2OSGAR @ transform @ OSGAR2CV

            # Our RANSAC calculated transform from key frame pose to current
            # frame pose, but we need the other direction.
            transform = inverse_rigid_transform(transform)

            # If we have an external estimate of the orientation of the camera,
            # we can correct the movement estimate.
            if camera_orientation is not None:
                # We are looking for a `correction` rotation such that after
                # we combine current motion estimate (`transform`) with previous
                # pose estimate, the new estimate's orientation matches
                # `camera_orientation`. I.e.:
                #   camera_orientation == key_frame_rotation @
                #                         correction @
                #                         transform_rotation
                inv_kf_pose_rot = inverse_rotation(self.key_frame_pose[:3, :3])
                inv_transform_rot = inverse_rotation(transform[:3, :3])
                assert (camera_orientation.shape == (3, 3))
                correction = as_rigid_transform(rot=(inv_kf_pose_rot @
                                                     camera_orientation @
                                                     inv_transform_rot))
                transform = correction @ transform

            # If we got too far from the key frame and we can switch to a new
            # key frame, we should do that. Otherwise it is safer to not believe
            # the estimate, because exceeding distance or rotation since key
            # frame when key frame is the *previous frame* (!) is very
            # suspicious.
            dist = np.linalg.norm(transform[:3, 3])
            # Take (1, 0, 0) vector, rotate it, measure the (cosine of) angle
            # between the new vector and the original (1, 0, 0) vector.
            rotation_value_cos = (transform[0, 0] /
                                  (1e-9 + np.linalg.norm(transform[0, :3])))
            if (dist > self.params.max_translation_since_key_frame
                    or rotation_value_cos < np.cos(
                        self.params.max_rotation_since_key_frame)):
                if self.can_advance_keyframe():
                    self.advance_keyframe()
                    return self.update(img, depth, camera_orientation,
                                       altitude, visualize)
                else:
                    if camera_orientation is not None:
                        # We can trust the rotation part, but not the
                        # translation part.
                        transform[:3, 3] = 0
                    else:
                        # We should not even trust the rotation part.
                        transform = no_change_rigid_transform()

            # New camera pose consists of transforming from current
            # camera coordinate frame to the coordinate frame of the key frame
            # (using `transform`) and from there to the coordinate system
            # centered where the camera started (using key_frame_pose).
            # Together, this makes a transform from "here" to "where camera
            # started", i.e. "current pose".
            self.trajectory.append(self.key_frame_pose @ transform)
        else:  # Oh no, RANSAC has not found anything!
            # Fallback to assuming constant velocity and rotation, i.e.
            # using `prev_rt` instead of `transform`.

            if camera_orientation is not None:
                # Similar calculation to the correction above, so see the
                # comment there.
                inv_pose_rot = inverse_rotation(self.pose[:3, :3])
                inv_prev_rot = inverse_rotation(prev_rt[:3, :3])
                correction = as_rigid_transform(
                    rot=(inv_pose_rot @ camera_orientation @ inv_prev_rot))
            else:
                correction = no_change_rigid_transform()

            # Similar reasoning as trajectory update above, except using
            # (possibly corrected for rotation) estimate from the previous
            # movement as our best guess.
            self.trajectory.append(self.pose @ correction @ prev_rt)

        # Make sure that despite all the chained matrix multiplications, the
        # rotation submatrix still represents rotation and does not include
        # scale and shear.
        self.pose[:3, :3] = near_rotation_matrix(self.pose[:3, :3])

        # It would be nice to apply some more advanced correction, like we can
        # do with rotation, but for now I don't see a generic one that is
        # guaranteed to exist. So let's do a simple correction.
        #
        # The difficult situation to correct is: Let's say that the estimated
        # movement (without any rotation) says dx,dy,dz==1,1,1. But altitude
        # change says that dz==4. What does it mean for dx and dy? That we
        # should quadruple them, to keep the proportions? Or that the robot
        # changed direction and most of the estimated travelled distance went
        # into dz and therefore dx and dy should be smaller? How much smaller,
        # if dz from atlimeter is a bigger change than the whole estimated
        # movement?
        if altitude is not None:
            self.pose[2, 3] = altitude

        self.prev_frame = frame

        if visualize:
            shown_img = cv2.cvtColor(frame.img, cv2.COLOR_GRAY2BGR)
            inls = np.zeros((curr_corners.shape[0], ), dtype=np.bool)
            inls[inliers] = True
            for kfc, cc, inl in zip(kf_corners, curr_corners, inls):
                # Inliers in green, outliers in red.
                color = (0, 255, 0) if inl else (0, 0, 255)
                kfc = tuple(kfc.astype(np.int))
                cc = tuple(cc.astype(np.int))
                cv2.circle(shown_img, cc, 2, color)
                cv2.line(shown_img, cc, kfc, color)
            cv2.imshow('Tracked features', shown_img)

        return self.pose


class RobotVisualOdometry:
    """Visual odometry in coordinate frame centered where robot started.

    You can use this class directly only if your global coordinate system
    aligns wit where the robot starts. Otherwise you should use
    `GlobalVisualOdometry`.
    """
    def __init__(self, camera_xyz, camera_rpy, odometry_params):
        """Initialize RobotVisualOdometry.

        camera_xyz -- XYZ triplet, in meters, representing position of the
                      camera relative to the center of the robot.
        camera_rpy -- Roll, pitch and yaw of the camera relative to the
                      orientation of the robot.
        odometry_params -- VisualOdometryParams.
        """
        self.camera_odometry = CameraVisualOdometry(odometry_params)

        self.camera_to_robot = as_rigid_transform(xyz=camera_xyz,
                                                  rpy=camera_rpy)
        self.robot_to_camera = inverse_rigid_transform(self.camera_to_robot)

        self.trajectory = [
            self.camera_to_robot @ self.camera_odometry.pose @
            self.robot_to_camera
        ]

    @property
    def xyz(self):
        """XYZ coordinates relative to initial position of the robot."""
        return np.asarray(self.pose)[:3, 3]

    @property
    def rpy(self):
        """Roll, pitch and yaw relative to initial orientation of the robot."""
        return rotation_matrix_to_euler_angles(self.pose[:3, :3])

    @property
    def pose(self):
        """Pose of the robot relative to its initial position.

        4x4 rigid transformation matrix capturing current rotation and
        translation of the robot relative to its initial position.

        I.e., this matrix transforms from coordinate system centered at the
        *current* robot location to a coordinate frame centered at the initial
        location of the robot.

        |R t|
        |0 1|
        """
        return self.trajectory[-1]

    def update(self, img, depth, imu_orientation=None, altitude=None,
               visualize=False):
        """Update the odometry estimate, extend `self.trajectory`.

        Depending on circumstances, key frame may get advanced.

        img -- Greyscale image.
        depth -- np.float32 depth map with the same resolution as `img`.
        imu_orientation -- Optional 3x3 rotation matrix representing
                           orientation *of the robot*, in the coordinate system
                           relative to robot's starting position, that we want
                           to get *after* the update finishes.
        altitude -- Optional altitude *of the robot*, in the coordinate system
                    relative to robot's starting position, that we want to get
                    *after* the update finishes.
        visualize -- If set to `True`, display an extra visualization with
                     detected image features and their tracking.
        """
        camera_orientation = (None if imu_orientation is None else
                              self.robot_to_camera[:3, :3] @ imu_orientation)
        camera_altitude = (None if altitude is None else
                           (self.robot_to_camera @
                            np.array([[0., 0., altitude, 1.]]).T)[2, 0])
        cam_pose = self.camera_odometry.update(img, depth, camera_orientation,
                                               camera_altitude, visualize)
        # To convert robot's center (0, 0, 0) in coordinate frame relative to
        # current position of the robot to a point relative to the initial
        # position of the robot, we need to do the following (reading from the
        # right):
        #  * Convert to the coordinate frame relative to current position of the
        #    camera.
        #  * Apply the odometry transform from camera that maps to coordinate
        #    frame centered at initial position of the camera.
        #  * Map that to the coordinate frame centered where the robot started.
        robot_pose = self.camera_to_robot @ cam_pose @ self.robot_to_camera
        self.trajectory.append(robot_pose)
        return robot_pose


class GlobalVisualOdometry:
    """Visual odometry in a given global coordinate frame.

    If your robot does not start at (0, 0, 0), heading in the direction of axis
    X, this is likely the class you want to use.
    """
    def __init__(self, initial_xyz, initial_rpy, camera_xyz, camera_rpy,
                 odometry_params):
        """Initialize GlobalVisualOdometry.

        initial_xyz -- Initial XYZ of the robot relative to global coordinate
                       coordinate frame, in meters.
        initial_rpy -- Initial roll, pitch and yaw of the robot relative to
                       global coordinate system, in radians.
        camera_xyz -- XYZ triplet, in meters, representing position of the
                      camera relative to the center of the robot.
        camera_rpy -- Roll, pitch and yaw of the camera relative to the
                      orientation of the robot.
        odometry_params -- VisualOdometryParams.
        """
        self.robot_odometry = RobotVisualOdometry(camera_xyz, camera_rpy,
                                                  odometry_params)

        self.robot_to_global = as_rigid_transform(xyz=initial_xyz,
                                                  rpy=initial_rpy)
        self.global_to_robot = inverse_rigid_transform(self.robot_to_global)

        self.trajectory = [self.robot_to_global @ self.robot_odometry.pose]

    @property
    def xyz(self):
        """XYZ coordinates of the robot in global coordinate system."""
        return np.asarray(self.pose)[:3, 3]

    @property
    def rpy(self):
        """Roll, pitch and yaw of the robot in global coordinate system."""
        return rotation_matrix_to_euler_angles(self.pose[:3, :3])

    @property
    def pose(self):
        """Pose of the robot in the global coordinate system.

        4x4 rigid transformation matrix capturing current rotation and
        translation in the fixed global coordinate system.

        I.e., this matrix transforms from coordinate system centered at the
        *current* robot location to a coordinate frame centered at *global*
        xyz=[0, 0, 0], rpy=[0, 0, 0].

        |R t|
        |0 1|
        """
        return self.trajectory[-1]

    def update(self, img, depth, imu_orientation=None, altitude=None,
               visualize=False):
        """Update the odometry estimate, extend `self.trajectory`.

        Depending on circumstances, key frame may get advanced.

        img -- Greyscale image.
        depth -- np.float32 depth map with the same resolution as `img`.
        imu_orientation -- Optional 3x3 rotation matrix representing
                           orientation *of the robot* in the *global* coordinate
                           system that we want to get *after* the update
                           finishes.
        altitude -- Optional altitude *of the robot* in the *global* coordinate
                    system that we want to get *after* the update finishes.
        visualize -- If set to `True`, display an extra visualization with
                     detected image features and their tracking.
        """
        robot_orientation = (None if imu_orientation is None else
                             self.global_to_robot[:3, :3] @ imu_orientation)
        robot_altitude = (None if altitude is None else
                         (self.global_to_robot @
                            np.array([0., 0., altitude, 1.]))[2, 0])
        robot_pose = self.robot_odometry.update(img, depth, robot_orientation,
                                               robot_altitude, visualize)
        # To convert robot's center (0, 0, 0) in coordinate frame relative to
        # current position of the robot to a point relative to the initial
        # position of the robot, we need to do the following (reading from the
        # right):
        #  * Convert to the coordinate frame relative to current position of the
        #    camera.
        #  * Apply the odometry transform from camera that maps to coordinate
        #    frame centered at initial position of the camera.
        #  * Map that to the coordinate frame centered where the robot started.
        global_pose = self.robot_to_global @ robot_pose
        self.trajectory.append(global_pose)
        return global_pose


class Localization(osgar.node.Node):
    def __init__(self, config, bus):
        super().__init__(config, bus)
        odometry_params = VisualOdometryParams(config['focal-length'],
                                               config['principal-point'],
                                               min_valid_depth=config.get('min-depth', 0.01),
                                               max_valid_depth=config.get('max-depth', 10.0))
        publish_pose3d = functools.partial(self.bus.publish, 'pose3d')
        self.demo = Demo(config['xyz'], np.radians(config['rpy']), odometry_params, callback=publish_pose3d)
        self.bus.register('pose3d')
        cv2.setNumThreads(1)

    def on_origin(self, data):
        if len(data) == 8:
            robot_name, x, y, z, qa, qb, qc, qd = data
            rpy = osgar.lib.quaternion.euler_zyx((qa, qb, qc, qd))[::-1]
            self.demo.on_initial_pose([x, y, z], rpy)

    def on_orientation(self, data):
        quaternion = data
        ypr = osgar.lib.quaternion.euler_zyx(quaternion)
        rpy = ypr[::-1]
        imu_rpy = rpy
        return self.demo.on_imu_rpy(imu_rpy)

    def on_image(self, data):
        grey_img = cv2.imdecode(np.frombuffer(data, np.uint8), cv2.IMREAD_GRAYSCALE)
        color_img = cv2.imdecode(np.frombuffer(data, np.uint8), cv2.IMREAD_COLOR)
        return self.demo.on_img(self.time.total_seconds(), grey_img, color_img)

    def on_depth(self, data):
        # Converting to meters.
        depth = (data * 1e-3).astype(np.float32)
        return self.demo.on_depth(self.time.total_seconds(), depth)

    def on_baseline(self, data):
        baseline_xyz, baseline_quaternion = data
        baseline_ypr = osgar.lib.quaternion.euler_zyx(baseline_quaternion)
        baseline_rpy = baseline_ypr[::-1]
        self.demo.on_baseline(baseline_xyz, baseline_rpy)

    def run(self):
        try:
            # wait for valid origin
            while True:
                dt, channel, data = self.bus.listen()
                if channel == "origin":
                    self.on_origin(data)
                    if self.demo.odometry is not None:
                        break

            while True:
                dt, channel, data = self.bus.listen()
                self.time = dt
                if channel == "orientation":
                    self.on_orientation(data)
                elif channel == "depth":
                    self.on_depth(data)
                elif channel == "image":
                    self.on_image(data)
                elif channel == "baseline":
                    self.on_baseline(data)
                elif channel == "origin":
                    pass
                else:
                    assert False, "unknown input channel"
        except osgar.bus.BusShutdownException:
            pass


class Demo:
    """Sample application and visualization of GlobalVisualOdometry."""
    def __init__(self, camera_xyz, camera_rpy, odometry_params, visualize=False,
                 output_trajectory_img_path=None, callback=None):
        """Initialize the demo.

        To complete the initialization, you also need to provide initial pose
        to on_initial_pose(). This can come from log or can be a fixed one.

        camera_xyz -- Placement of the camera on the robot.
        camera_rpy -- Roll, pitch and yaw (radians) of camera orientation on the
                      robot.
        odometry_params -- VisualOdometryParams()
        callback -- is called whenever a new estimate is available
        visualize -- If True, show extra visualizations.
        output_trajectory_img_path -- Path to an image file where trajectory
                                      should be stored.
        """
        self.camera_xyz = camera_xyz
        self.camera_rpy = camera_rpy
        self.odometry_params = odometry_params
        self.odometry = None
        self.callback = callback
        self.visualize = visualize
        self.output_trajectory_img_path = output_trajectory_img_path

        # Information about sensor measurements.
        self.depth = None
        self.depth_time = None
        self.grey_img = None
        self.color_img = None
        self.img_time = None
        self.baseline_xyz = None
        self.baseline_rpy = None
        self.baseline_trajectory_xy = []
        self.altitude = None
        self.imu_rpy = None

        # For visualization. Minimum and maximum X and Y values observed in the
        # trajectory. They are merged into the same variables to keep the
        # scaling ratio same for both axes when visualizing the trajectory.
        self.min_xy = self.max_xy = 0
        self._waitKey_timeout = 0

    def on_initial_pose(self, initial_xyz, initial_rpy):
        """Finish initialization by setting global coordinate frame."""
        self.odometry = GlobalVisualOdometry(initial_xyz, initial_rpy,
                                             self.camera_xyz, self.camera_rpy,
                                             self.odometry_params)
        # To keep baseline_trajectory same length as odometry.trajectory, so
        # that pairwise comparison matches related positions.
        initial_pose = as_rigid_transform(xyz=initial_xyz, rpy=initial_rpy)
        self.baseline_trajectory_xy = [np.asarray(
            initial_pose[:2,3]).tolist()]

    def __maybe_update_odometry_and_visualize(self):
        """Update odometry with a matching accumulated image-depth pair.

        If initial pose is not yet known or there is no matching accumulated
        image-depth pair available, do nothing.

        Return True if awaiting more data, False if user requested a stop.
        """
        if self.odometry is None:
            # This can happen before the first information about origin arrives
            # and we need to wait a bit more.
            return True

        if (self.grey_img is None or self.depth is None
                or abs(self.depth_time - self.img_time) > 0.15):
            return True  # Nothing to do now, but that's OK.

        imu_orientation = (None if self.imu_rpy is None else
                           euler_angles_to_rotation_matrix(*self.imu_rpy))
        global_pose = self.odometry.update(self.grey_img, self.depth,
                                           imu_orientation, self.altitude,
                                           self.visualize)
        xyz = np.asarray(global_pose)[:3, 3]
        rpy = rotation_matrix_to_euler_angles(global_pose[:3, :3])
        quat = osgar.lib.quaternion.from_rotation_matrix(global_pose[:3, :3])
        if self.callback is not None:
            # `tolist` changes numpy scalars to Python scalars
            self.callback([xyz.tolist(), quat])

        # User interface, ha ha.
        g_logger.info('Estimate: xyz={}, rpy={}'.format(xyz, np.degrees(rpy)))
        if self.baseline_xyz is not None:
            drift = np.sqrt(np.sum((xyz - np.asarray(self.baseline_xyz))**2))
            g_logger.info('Baseline: xyz={}, rpy={}'.format(
                self.baseline_xyz, np.degrees(self.baseline_rpy)))
            g_logger.info(f'Drift: {drift}')

        if self.baseline_xyz is not None:
            self.baseline_trajectory_xy.append(self.baseline_xyz[:2])
            self.min_xy = min(self.min_xy, min(*self.baseline_xyz[:2]))
            self.max_xy = max(self.max_xy, max(*self.baseline_xyz[:2]))

        self.min_xy = min(self.min_xy, min(*xyz[:2]))
        self.max_xy = max(self.max_xy, max(*xyz[:2]))

        if self.visualize or self.output_trajectory_img_path:
            trajectory_img = np.zeros((1000, 1000, 3), dtype=np.uint8)

            def img_coord(wx, wy):
                MARGIN = 1
                ix = int(trajectory_img.shape[1] * (wx -
                                                    (self.min_xy - MARGIN)) /
                         ((self.max_xy + MARGIN) - (self.min_xy - MARGIN)))
                iy = int(trajectory_img.shape[0] * (wy -
                                                    (self.min_xy - MARGIN)) /
                         ((self.max_xy + MARGIN) - (self.min_xy - MARGIN)))
                return ix, trajectory_img.shape[0] - MARGIN - iy

            #assert(len(self.baseline_trajectory_xy) == len(self.odometry.trajectory))
            for bxy_a, bxy_b in zip(self.baseline_trajectory_xy[:-2],
                                    self.baseline_trajectory_xy[1:]):
                cv2.line(trajectory_img, img_coord(*bxy_a), img_coord(*bxy_b),
                         (0, 255, 0))
            for g_pose_a, g_pose_b in zip(self.odometry.trajectory[:-2],
                                          self.odometry.trajectory[1:]):
                cv2.line(trajectory_img,
                         img_coord(g_pose_a[0, 3], g_pose_a[1, 3]),
                         img_coord(g_pose_b[0, 3], g_pose_b[1, 3]),
                         (0, 255, 255))
            cv2.imshow(
                'Trajectory: green - baseline, yellow - visual ' + 'odometry',
                trajectory_img)
            if self.output_trajectory_img_path:
                cv2.imwrite(self.output_trajectory_img_path, trajectory_img)

            if self.visualize:
                cv2.imshow('img', self.color_img)
                KEY_Q = ord('q')
                KEY_SPACE = ord(' ')
                key = cv2.waitKey(self._waitKey_timeout) & 0xFF
                if key == KEY_Q:
                    return False
                if key == KEY_SPACE:
                    self._waitKey_timeout = 0 if self._waitKey_timeout > 0 else 1

        # Discard measurements, because they are now obsolete ("consumed.")
        self.imu_rpy = None
        self.altitude = None
        self.grey_img = None
        self.color_img = None
        self.depth = None

        return True

    def on_img(self, timestamp, grey_img, color_img):
        """Accumulate image information and maybe update odometry.

        Update odometry and visualize in case there is a matching image-depth
        pair available now.

        timestamp -- current time in seconds
        grey_img -- Greyscale image used for visual odometry.
        color_img -- color image used for visualization to the user.

        Return True if awaiting more data, False if user requested a stop.
        """
        self.img_time = timestamp
        self.grey_img = grey_img
        self.color_img = color_img
        return self.__maybe_update_odometry_and_visualize()

    def on_depth(self, timestamp, depth):
        """Accumulate depth information and maybe update odometry.

        Update odometry and visualize in case there is a matching image-depth
        pair available now.

        timestamp -- current time in seconds
        depth -- np.float32 depth map in meters

        Return True if awaiting more data, False if user requested a stop.
        """
        self.depth_time = timestamp
        self.depth = depth
        return self.__maybe_update_odometry_and_visualize()

    def on_imu_rpy(self, imu_rpy):
        """Remember measurement from IMU."""
        self.imu_rpy = imu_rpy

    def on_altitude(self, altitude):
        """Remember measurement from altimeter."""
        self.altitude = altitude

    def on_baseline(self, baseline_xyz, baseline_rpy):
        """Remember baseline trajectory."""
        self.baseline_xyz = baseline_xyz
        self.baseline_rpy = baseline_rpy


def main():
    cv2.setNumThreads(1)
    import argparse
    import json

    from osgar.lib.quaternion import euler_zyx
    from osgar.lib.serialize import deserialize
    from osgar.logger import LogReader, lookup_stream_id

    parser = argparse.ArgumentParser()
    parser.add_argument('--logfile', help='Path to Osgar log.')
    parser.add_argument('--origin-stream',
                        help='Stream id or name with data about robot\'s ' +
                        'coordinates in the starting area.')
    parser.add_argument('--image-stream',
                        help='Stream id or name with image data.')
    parser.add_argument('--depth-stream',
                        help='Stream id or name with depth data.')
    parser.add_argument('--imu-stream',
                        help='Optional stream id or name with imu data.')
    parser.add_argument('--altitude-stream',
                        help='Optional stream id or name with altitude data.')
    parser.add_argument(
        '--baseline-3d-stream',
        help='Optional stream id or name with 3d coordinates ' +
        'to compare to.')
    parser.add_argument('--camera-focal-length',
                        type=float,
                        default=462.1,
                        help='Focal length of the camera in pixels.')
    parser.add_argument('--camera-center',
                        type=float,
                        nargs=2,
                        default=[319.5, 179.5],
                        help='X and Y coordinate of focal point in pixels.')
    parser.add_argument('--camera-xyz',
                        type=float,
                        nargs=3,
                        default=[0.23, 0, 0.19 + 0.06256005],
                        help='Placement of camera on the robot.')
    parser.add_argument('--camera-rpy',
                        type=float,
                        nargs=3,
                        default=[0., 0., 0.],
                        help='Heading of camera on the robot. Roll, pitch ' +
                        'and yaw in degrees.')
    parser.add_argument(
        '--min-depth',
        type=float,
        default=0.01,
        help='Measurements below this distance in meters are ' + 'ignored.')
    parser.add_argument('--max-depth',
                        type=float,
                        default=10.0,
                        help='Measurements beyond this distance in meters ' +
                        'are ignored.')
    parser.add_argument('--other-odometry-params',
                        type=json.loads,
                        default={},
                        help='Other VisualOdometryParams parameters in JSON ' +
                        'fromat.')
    parser.add_argument(
        '--trajectory-image-path',
        type=str,
        default=None,
        help='Where to store a generated image with baseline ' +
        'and estimated trajectory.')
    parser.add_argument(
        '--show',
        default=False,
        action='store_true',
        help='Enable visualization of the scene and detected ' + 'artifacts.')
    args = parser.parse_args()
    assert (args.logfile)
    assert (args.image_stream)
    assert (args.depth_stream)
    assert (args.max_depth > 0)
    assert (args.min_depth >= 0)
    assert (args.max_depth > args.min_depth)

    odometry_params = VisualOdometryParams(args.camera_focal_length,
                                           args.camera_center,
                                           min_valid_depth=args.min_depth,
                                           max_valid_depth=args.max_depth,
                                           **args.other_odometry_params)
    demo = Demo(args.camera_xyz, np.radians(args.camera_rpy), odometry_params,
                args.show, args.trajectory_image_path)

    image_stream_id = lookup_stream_id(args.logfile, args.image_stream)
    depth_stream_id = lookup_stream_id(args.logfile, args.depth_stream)
    streams = [image_stream_id, depth_stream_id]
    if args.origin_stream:
        origin_stream_id = lookup_stream_id(args.logfile, args.origin_stream)
        streams.append(origin_stream_id)
    else:
        origin_stream_id = None
        demo.on_initial_pose([0., 0., 0.], [0., 0., 0.])
    if args.imu_stream:
        imu_stream_id = lookup_stream_id(args.logfile, args.imu_stream)
        streams.append(imu_stream_id)
    else:
        imu_stream_id = None
    if args.altitude_stream:
        altitude_stream_id = lookup_stream_id(args.logfile,
                                              args.altitude_stream)
        streams.append(altitude_stream_id)
    else:
        altitude_stream_id = None
    if args.baseline_3d_stream:
        baseline_3d_stream_id = lookup_stream_id(args.logfile,
                                                 args.baseline_3d_stream)
        streams.append(baseline_3d_stream_id)
    else:
        baseline_3d_stream_id = None

    with LogReader(args.logfile, only_stream_id=streams) as logreader:
        for time, stream, data in logreader:
            data = deserialize(data)
            now = time.total_seconds()

            # In practice, it probably makes lots of sense to keep
            # re-initializing visual odometry as long as information about
            # origin keeps coming. The initial part, when the robot is in a
            # featureless wide area is difficult for visual odometry anyway.
            # Re-initializing would make odometry really start estimating
            # only once the robot enters competition area, which should,
            # presumably, be visually more diverse.
            if stream == origin_stream_id and demo.odometry is None:
                robot_name, x, y, z, qa, qb, qc, qd = data
                rpy = euler_zyx((qa, qb, qc, qd))[::-1]
                demo.on_initial_pose([x, y, z], rpy)
            elif stream == image_stream_id:
                grey_img = cv2.imdecode(np.frombuffer(data, np.uint8),
                                        cv2.IMREAD_GRAYSCALE)
                color_img = cv2.imdecode(np.frombuffer(data, np.uint8),
                                         cv2.IMREAD_COLOR)
                if not demo.on_img(now, grey_img, color_img):
                    break
            elif stream == depth_stream_id:
                # Converting to meters.
                depth = (data * 1e-3).astype(np.float32)
                if not demo.on_depth(now, depth):
                    break
            elif stream == imu_stream_id:
                quaternion = data
                ypr = euler_zyx(quaternion)
                rpy = ypr[::-1]
                imu_rpy = rpy
                demo.on_imu_rpy(imu_rpy)
            elif stream == altitude_stream_id:
                # TODO: Assuming one float in meters. Needs to be converted if
                #       we get, for example, millimeters instead.
                altitude = data
                demo.on_altitude(altitude)
            elif stream == baseline_3d_stream_id:
                baseline_xyz, baseline_quaternion = data
                baseline_ypr = euler_zyx(baseline_quaternion)
                baseline_rpy = baseline_ypr[::-1]
                demo.on_baseline(baseline_xyz, baseline_rpy)


if __name__ == '__main__':
    main()
